 /*
  * filereq.c - GamePark 32 file requester
  * Copyright (c)2002 Christian Nowak <chnowak@web.de>
  *
  *
  * This file is part of PlusIV.
  *
  * PlusIV is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * PlusIV is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with Foobar; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "defines.h"
#include "gpstdlib.h"
#include "gpgraphic.h"
#include "gpdef.h"
#include "gpstdio.h"
#include "gpfont.h"
#include "filereq.h"

#define FSEL_MIDDLE 7

typedef struct FINFO
  {
    char name[42];
    int size;
    BOOL isdir;
  } FINFO;

static BOOL FileIsDir ( char * fname )
  {
    GPFILEATTR attr;

    if (GpFileAttr ( fname, &attr )!=SM_OK)
      return FALSE;

    return ((attr.attr&0x10)==0x10);
  }

static void DisplayEntries ( FSEL * fsel, FINFO ** pdirlist, int num_entries, int index )
  {
    int i;
    int screen_start,   /* x/8 position on screen to start rendering */
        index_start;    /* Index position in directory to start */

    if (index>=num_entries)
      index = 0;

    if (index>FSEL_MIDDLE)
      {
        index_start = index-FSEL_MIDDLE;
        screen_start = 0;
      }
    else
      {
        index_start = 0;
        screen_start = FSEL_MIDDLE-index;
      }

    for (i=(320*240)-1;i>=0;--i)
      fsel->gpDraw->ptbuffer[i] = fsel->bgcolor;
    for (i=0;(i+screen_start)<14 && (i+index_start)<num_entries;i++)
      {
        u8 col;
        if ( pdirlist[index_start+i]->isdir )
          {
            /* Directory */
            sprintf ( fsel->fname, "<%s>", pdirlist[index_start+i]->name );
            col = (i+index_start)==index ? fsel->dirselcolor : fsel->dircolor;
          } else
          {
            /* Ordinary file */
            int j,k,l;
            char str[42];

            strcpy ( fsel->fname, pdirlist[index_start+i]->name );
            sprintf(str,"%d",pdirlist[index_start+i]->size);
            l = strlen(fsel->fname);
            k = 40-l-strlen(str);
            for (j=strlen(fsel->fname);j<k+l;j++)
              fsel->fname[j] = ' ';
            fsel->fname[j] = '\0';
            strcat(fsel->fname, str);

            col = (i+index_start)==index ? fsel->fileselcolor : fsel->filecolor;
          }
        GpTextOut(NULL, fsel->gpDraw, 0, (1+i+screen_start)*16, fsel->fname, col);
      }
  }

static void sortDirEntries ( FINFO ** pdirlist, int num )
  {
    int i,j;

    if (num<2)
      return;

    for (i=0;i<num;i++)
      {
        for (j=0;j<num-1;j++)
          {
            if (strcmp(pdirlist[j]->name, pdirlist[j+1]->name)>0)
              {
                FINFO * temp;
                temp = pdirlist[j];
                pdirlist[j] = pdirlist[j+1];
                pdirlist[j+1] = temp;
              }
          }
      }
  }

static int SelectFile ( FSEL * fsel, char * basedir, BOOL enableCancel )
  {
    unsigned int i;
    int j;
    int numdirs = 0;
    ulong num_entries;
    ulong read_count;
    int index = 0;
    int lk = 0,k;
    FINFO * dirlist;
    FINFO ** pdirlist;
    GPDIRENTRY * direntry;
    char p[256];

    for (i=0;i<320*240;++i)
      fsel->gpDraw->ptbuffer[i] = fsel->bgcolor;
    GpTextOut(NULL, fsel->gpDraw, 80, 116, "Scanning directory...", fsel->filecolor);

    /* Get directory */
    if ( GpDirEnumNum(basedir, &num_entries)!=SM_OK )
      return -1;
    if (num_entries==0)
      return -1;

    dirlist = malloc ( num_entries*sizeof(FINFO) );
    memset ( dirlist, 0, sizeof(FINFO)*num_entries );
    pdirlist = malloc ( num_entries*sizeof(FINFO*) );

    direntry = malloc ( sizeof(GPDIRENTRY)*num_entries );
    GpDirEnumList ( basedir, 0, num_entries, direntry, &read_count );
    num_entries = read_count;

    for (i=0;i<num_entries;i++)
      {
        strcpy ( dirlist[i].name, direntry[i].name );
        if ( (strcmp("." , direntry[i].name)==0) ||
             (strcmp("..", direntry[i].name)==0) ||
             FileIsDir(direntry[i].name) )
          {
            dirlist[i].isdir = TRUE;
            numdirs++;
            dirlist[i].size = -1;
          }
        else
          {
            ulong s = 0;
            GpFileGetSize ( direntry[i].name, &s );
            dirlist[i].isdir = FALSE;
            dirlist[i].size = (int)s;
          }
        pdirlist[i] = &dirlist[i];
      }

    free ( direntry );

    /* Separate dirs from files */
    j=0;i=0;
    while (pdirlist[j]->isdir && j<num_entries) j++;
    for (i=j;i<num_entries;i++)
      {
        if (pdirlist[i]->isdir)
          {
            FINFO * tmp = pdirlist[i];
            pdirlist[i] = pdirlist[j];
            pdirlist[j] = tmp;
            j++;
          }
      }
    /* Sort directories */
    sortDirEntries ( pdirlist, numdirs );
    /* Sort files */
    sortDirEntries ( &pdirlist[numdirs], num_entries-numdirs );

    DisplayEntries ( fsel, pdirlist, num_entries, index );
    GpRelativePathGet(p);
    GpTextOut(NULL,fsel->gpDraw,0,0,p,fsel->curdircolor);
    while (GpKeyGet()&GPC_VK_FA); /* Wait until the A button is released */
    while (!((k=GpKeyGet())&(GPC_VK_FA|(enableCancel?GPC_VK_FB:0))) )
      {
        k &= GPC_VK_UP | GPC_VK_DOWN | GPC_VK_LEFT | GPC_VK_RIGHT;

        if (lk==0)
          {
            if (k&GPC_VK_UP)
              index--;
            else
            if (k&GPC_VK_DOWN)
              index++;
            else
            if (k&GPC_VK_LEFT)
              index-=15;
            else
            if (k&GPC_VK_RIGHT)
              index+=15;

            if (index<0) index=0;
            if (index>=(int)num_entries) index=num_entries-1;

            /* Display entries */
            if (k&(GPC_VK_UP|GPC_VK_DOWN|GPC_VK_LEFT|GPC_VK_RIGHT))
              {
                DisplayEntries ( fsel, pdirlist, num_entries, index );
                GpRelativePathGet(p);
                GpTextOut(NULL,fsel->gpDraw,0,0,p,fsel->curdircolor);
              }
          }

        lk = k;
      }

    strcpy(fsel->fname, pdirlist[index]->name);
    free ( pdirlist );
    free ( dirlist );

    if ( enableCancel )
      return k&GPC_VK_FB ? -1 : 0;
    else
      return 0;
  }

int FileRequest ( FSEL * fsel, char * bdir, BOOL enableCancel )
  {
    char * basedir;
    BOOL isdir = FALSE;
    
    basedir=bdir;
    if (!FileIsDir(basedir))
      {
        basedir="gp:\\gpmm";
        if (!FileIsDir(basedir))
          {
            basedir="gp:\\";
          }
      }
    GpRelativePathSet ( basedir );

    do
      {
        char path[256];
        if (SelectFile(fsel, "\\", enableCancel)!=0)
          return -1;
        GpRelativePathGet(path);
        if (strcmp(fsel->fname,"..")==0)
          {
            int i;
            int l=strlen(path);
            path[l-1]='0';
            for (i=l-2;i>0 && path[i]!='\\';i--)
              path[i]='\0';
            isdir=TRUE;
          } else
        if (strcmp(fsel->fname,".")==0)
          {
            isdir=TRUE;
          } else
          {
            if ((isdir=FileIsDir(fsel->fname)))
              strcat(path,fsel->fname);
          }

        if (isdir)
          GpRelativePathSet(path);
        GpRelativePathGet(path);
      } while (isdir);
    return 0;
  }
