#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "defines.h"
#include "gpsoundbuf.h"
#include "../../p4/p4.h"
#include "kboard.h"
#include "filereq.h"
#include "gpmenu.h"

#include "gpgraphic.h"
#include "gpdef.h"
#include "gpstdlib.h"

#include "gfx.h"

extern void SetPlus4Palette ( u32 );
extern u32 brightness;
extern u8 kernal[];
extern GPSOUNDBUF sb;
extern int menuselection;

u8 arrow[8*8] =
  {
    2,2,2,2,1,2,2,2,
    2,2,2,2,2,1,2,2,
    2,2,2,2,2,2,1,2,
    1,1,1,1,1,1,1,1,
    2,2,2,2,2,2,1,2,
    2,2,2,2,2,1,2,2,
    2,2,2,2,1,2,2,2,
    2,2,2,2,2,2,2,2
  };

u8 kbcodes[5][18] =
  {
    {
      0x64, /* ESC */
      0x70, /* 1 */
      0x73, /* 2 */
      0x10, /* 3 */
      0x13, /* 4 */
      0x20, /* 5 */
      0x23, /* 6 */
      0x30, /* 7 */
      0x33, /* 8 */
      0x40, /* 9 */
      0x43, /* 0 */
      0x66, /* + */
      0x56, /* - */
      0x65, /* = */
      0xff,
      0xff,
      0xff,
      0x04  /* F1 */
    },
    {
      0xff,
      0x76, /* Q */
      0x11, /* W */
      0x16, /* E */
      0x21, /* R */
      0x26, /* T */
      0x31, /* Y */
      0x36, /* U */
      0x41, /* I */
      0x46, /* O */
      0x51, /* P */
      0x07, /* @ */
      0xff,
      0x61, /* * */
      0xff,
      0xff,
      0xff,
      0x05  /* F2 */
    },
    {
      0xff,
      0x12, /* A */
      0x15, /* S */
      0x22, /* D */
      0x25, /* F */
      0x32, /* G */
      0x35, /* H */
      0x42, /* J */
      0x45, /* K */
      0x52, /* L */
      0xff,
      0xff,
      0xff,
      0x01, /* Return */
      0xff,
      0xff,
      0xff,
      0x06  /* F3 */
    },
    {
      0x75, /* C= */
      0x14, /* Z */
      0x27, /* X */
      0x24, /* C */
      0x37, /* V */
      0x34, /* B */
      0x47, /* N */
      0x44, /* M */
      0x57, /* < */
      0x54, /* > */
      0xff, /* ? */
      0xff,
      0xff,
      0xff,
      0xff,
      0xff,
      0xff,
      0xff
    },
    {
      /* Space */
      0x74,
      0x74,
      0x74,
      0x74,
      0x74,
      0x74,
      0x74,
      0x74,
      0x74,
      0x74,
      0x74,
      0x74,
      0x74,
      0x74,
      0x74,
      0x74,
      0x74,
      0x74
    }
  };

/* From D64.c */
extern char * d64_ftypes[8];

int kbx=0,kby=0;
int mode=0;   /* 0 - Keyboard
                 1 - Joy 1
                 2 - Joy 2
                 3 - Cursor
              */
BOOL prevM = FALSE;
BOOL prevL = FALSE;
extern u8 * screenbuf;
extern GPDRAWSURFACE gpDraw[2];

extern void XlatVGABuffer (unsigned char *src, unsigned char *dest);

/* Uses the Commodore font at kernal[0x1000] to print
   a character into the VGA buffer */
void Commo_CharOut ( u8 * buffer, char d, unsigned int x, unsigned int y, u8 fgcolor, u8 bgcolor, BOOL reverse )
  {
    unsigned int i,j;
    unsigned int fontoffs;
    u8 c = d;
    u8 * b = &buffer[(320*y)+x];

    if (reverse)
      {
        u8 temp;

        temp = fgcolor;
        fgcolor = bgcolor;
        bgcolor = temp;
      }

    if ( c>=192 && c<=223 )
      c -= 96;
    else if ( c>=224 && c<=254 )
      c -= 64;
    else if ( c==255 )
      c = 126;

    if (c>=64 && c<=95)
      c-=64;
    else if (c>=96 && c<=128)
      c-=32;
    else if (c>=149 && c<=155)
      c-=64;
    else if (c>=161 && c<=191)
      c-=64;

    fontoffs = ((unsigned int)c)<<3;

    for (i=j=0;i<8;i++,j+=320)
      {
        u8 cl = kernal[0x1000+fontoffs+i];
        b[j+0] = cl & 0x80 ? fgcolor:bgcolor;
        b[j+1] = cl & 0x40 ? fgcolor:bgcolor;
        b[j+2] = cl & 0x20 ? fgcolor:bgcolor;
        b[j+3] = cl & 0x10 ? fgcolor:bgcolor;
        b[j+4] = cl & 0x08 ? fgcolor:bgcolor;
        b[j+5] = cl & 0x04 ? fgcolor:bgcolor;
        b[j+6] = cl & 0x02 ? fgcolor:bgcolor;
        b[j+7] = cl & 0x01 ? fgcolor:bgcolor;
      }
  }

void Commo_TextOut ( u8 * buffer, char * t, int tlen, unsigned int x, unsigned int y, u8 fgcolor, u8 bgcolor, BOOL inv )
  {
    int i;

    for (i=0;i<tlen;i++)
      Commo_CharOut ( buffer, t[i], x+(i<<3), y, fgcolor, bgcolor, inv );
  }

/* Displays a message on the screen and waits until the A button is pressed */
void Message ( GPDRAWSURFACE * gpDraw, char * m, u8 bgcolor, u8 color )
  {
    int i;

    for (i=0;i<(320*240);i++)
      gpDraw->ptbuffer[i] = bgcolor;
    GpTextOut(NULL, gpDraw, (320-(strlen(m)<<3))>>1, 120-8, m, color);
    GpTextOut(NULL, gpDraw, (320-(14<<3))>>1, 120+16, "Press A Button", color);

    while (GpKeyGet()!=0);
    while (!(GpKeyGet()&GPC_VK_FA));
  }

void SetKbVal ( P4 * machine, int kbx, int kby, u8 val )
  {
    u8 c = kbcodes[kby][kbx];
    u8 nx,ny;

    if (c==0xff)
      return;

    nx = (c>>4)&7;
    ny = c&7;
    machine->keys[nx][ny] = val;
  }

void Inverse ( int x, int y )
  {
    int sx = 88+(x*8),
        sy = 320*(200+(y*8)),
        i,j;

    for (j=0;j<8;sy+=320,j++)
      for (i=0;i<8;i++)
        screenbuf[sx+sy+i] ^= 3;
  }

void PaintArrow ( void )
  {
    int i,j;

    /* Erase */
    for (j=320*200;j<(320*200)+(4*8*320);j+=320)
      for (i=0;i<8;i++)
        screenbuf[i+j] = 0x82;

    /* Paint */
    for (i=0;i<8;i++)
      for (j=0;j<8;j++)
        screenbuf[(320*(i+200+(mode*8)))+j] = arrow[j+(8*i)] | 0x80;
  }

void InitKb ( P4 * machine )
  {
    int i;
    for (i=(320*200);i<(320*240);i++)
      {
        screenbuf[i] = kboard[i-(320*200)]+0x80;
      }
    Inverse ( kbx, kby );
    PaintArrow();
  }

int LoadPRG ( P4 * machine )
  {
    FSEL fsel;
    int l;
    char fname[256];

    fsel.bgcolor = 0x82;
    fsel.dircolor = 0x84;
    fsel.filecolor = 0x80;
    fsel.dirselcolor = 0x81;
    fsel.fileselcolor = 0x81;
    fsel.curdircolor = 0x81;
    fsel.fname = fname;
    fsel.gpDraw = &gpDraw[0];
    if (FileRequest ( &fsel, "gp:\\gpmm\\p4emu", TRUE )==0)
      {
        l=strlen(fsel.fname);
        if (l>4)
          {
            if ((tolower(fsel.fname[l-4])=='.') &&
                (tolower(fsel.fname[l-3])=='p') &&
                (tolower(fsel.fname[l-2])=='r') &&
                (tolower(fsel.fname[l-1])=='g'))
              {
                if (P4_LoadPRG(machine, fname)!=0)
                  {
                    Message ( &gpDraw[0], "Loading of the PRG failed !", 0x82, 0x80 );
                  } else
                  {
                    char str[64];
                    sprintf ( str, "Loaded \"%s\"", fname );
                    Message ( &gpDraw[0], str, 0x82, 0x80 );
                    return 1;
                  }
              } else
              {
                Message ( &gpDraw[0], "Select a file with the .PRG suffix", 0x82, 0x80 );
              }
          }
      } else
      {
        Message ( &gpDraw[0], "Loading cancelled", 0x82, 0x80 );
      }
    return 0;
  }

int GetNumD64Entries ( P4 * machine )
  {
    int i;
    D64 * disk = &(((D1551*)(machine->drive))->disk);

    if (D64_GotoFirstDirEntry(disk)!=0)
      return 0;

    for (i=1;D64_GotoNextDirEntry(disk)==0;i++);
    return i;
  }

int ShowD64Entries ( P4 * machine, int numentries, int index, char * fname )
  {
    char str[64];
    D64 * disk = &(((D1551*)(machine->drive))->disk);
    int i;
    int screen_start;
    int index_start;

    if (index>=numentries)
      index=0;

    if (index>14)
      {
        index_start = index-14;
        screen_start = 0;
      } else
      {
        index_start = 0;
        screen_start = 14-index;
      }

    memset ( screenbuf, 0x82, 320*240 );

    /* Show disk name */
    D64_GetDiskName ( disk, str );
    Commo_TextOut ( screenbuf, str, 16, 0, 0, 0x80, 0x82, TRUE );

    /* Show dir entries */
    for (i=0;(i+screen_start)<29 && (i+index_start)<numentries;i++)
      {
        int res;
        char ts[64];

        if (i==0)
          {
            int j;

            res = D64_GotoFirstDirEntry(disk);
            if (index_start>0)
              {
                for (j=0;j<index_start;j++)
                  res = D64_GotoNextDirEntry(disk);
              }
          } else
          {
            res = D64_GotoNextDirEntry(disk);
          }

        if (res==0)
          {
            int j;

            memset ( str, 0x20, 40 );
            D64_GetEntryName(disk,ts);
            sprintf(&str[00], "%d", D64_GetEntrySize(disk));
            sprintf(&str[05], "\"%s\"", ts);
            sprintf(&str[05+19], "%s", d64_ftypes[disk->direntry->file_type&0x07]);
            for (j=0;j<40;j++)
              if (str[j]=='\0') str[j]=' ';
            str[40] = '\0';
            Commo_TextOut ( screenbuf, str, 40, 0, (1+i+screen_start)*8, (i+index_start)==index?0x81:0x80, 0x82, (i+index_start)==index );
            if ( (i+index_start)==index )
              strcpy ( fname, ts );
          }
      }

    XlatVGABuffer ( screenbuf, gpDraw[0].ptbuffer );

    return 0;
  }

int SelectD64File ( P4 * machine, char * fname )
  {
    u16 k;
    int index = 0;
    int numentries = GetNumD64Entries ( machine );

    while (1)
      {
        ShowD64Entries ( machine, numentries, index, fname );
        while (GpKeyGet());
        while (!(k=GpKeyGet()&(GPC_VK_FA|GPC_VK_FB|GPC_VK_UP|GPC_VK_DOWN|GPC_VK_LEFT|GPC_VK_RIGHT)));

        if ( k&GPC_VK_FB )
          return -1;
        else if ( k&GPC_VK_FA )
          return index;
        else if ( k&GPC_VK_UP )
          index--;
        else if ( k&GPC_VK_DOWN )
          index++;
        else if ( k&GPC_VK_LEFT )
          index-=14;
        else if ( k&GPC_VK_RIGHT )
          index+=14;

        if ( index<0 )
          index = 0;
        else if (index>=numentries)
          index = numentries-1;
      }
  }

int menucallback ( GPMENU * menu, u16 k )
  {
    P4 * machine = (P4*) menu->userdata;

    switch (menu->selection)
      {
        case 0: /* Attach D64 image */
          {
            FSEL fsel;
            int l;
            char fname[256];

            fsel.bgcolor = 0x82;
            fsel.dircolor = 0x84;
            fsel.filecolor = 0x80;
            fsel.dirselcolor = 0x81;
            fsel.fileselcolor = 0x81;
            fsel.curdircolor = 0x81;
            fsel.fname = fname;
            fsel.gpDraw = &gpDraw[0];
            if (FileRequest ( &fsel, "gp:\\gpmm\\p4emu", TRUE )==0)
              {
                l=strlen(fsel.fname);
                if (l>4)
                  {
                    if ((tolower(fsel.fname[l-4])=='.') &&
                        (tolower(fsel.fname[l-3])=='d') &&
                        (tolower(fsel.fname[l-2])=='6') &&
                        (tolower(fsel.fname[l-1])=='4'))
                      {
                        D64_UnloadImage ( &(((D1551*)(machine->drive))->disk) );
                        if (D64_LoadImage ( &(((D1551*)(machine->drive))->disk), fname)!=0)
                          {
                            Message ( &gpDraw[0], "Loading of the D64 failed !", 0x82, 0x80 );
                          } else
                          {
                            char str[64];
                            sprintf ( str, "Attached \"%s\"", fname );
                            Message ( &gpDraw[0], str, 0x82, 0x80 );
                          }
                      } else
                      {
                        Message ( &gpDraw[0], "Select a file with the .D64 suffix", 0x82, 0x80 );
                      }
                  }
              } else
              {
                Message ( &gpDraw[0], "Loading cancelled", 0x82, 0x80 );
              }
            }
          break;
        case 1: /* Load PRG from D64 image */
          if (((D1551*)(machine->drive))->disk.image == NULL)
            {
              Message ( &gpDraw[0], "Attach a D64 image first !", 0x82, 0x80 );
            }
          else
            {
              int sel;
              char fname[32];

              if ((sel=SelectD64File ( machine, fname ))==-1)
                {
                  Message ( &gpDraw[0], "Loading cancelled", 0x82, 0x80 );
                } else
                {
                  int j;
                  u16 l = 0;

                  machine->RAM[0x0527+l++] = 76;  /* L */
                  machine->RAM[0x0527+l++] = 79;  /* O */
                  machine->RAM[0x0527+l++] = 65;  /* A */
                  machine->RAM[0x0527+l++] = 68;  /* D */
                  machine->RAM[0x0527+l++] = 34;  /* " */
                  for (j=0;j<strlen(fname);j++)
                    machine->RAM[0x0527+l++] = fname[j];
                  machine->RAM[0x0527+l++] = 34;  /* " */
                  machine->RAM[0x0527+l++] = 44;  /* , */
                  machine->RAM[0x0527+l++] = 56;  /* 8 */
                  machine->RAM[0x0527+l++] = 13;
                  machine->RAM[0x00ef] = l;
                  return 1;
                }
            }
          break;
        case 2: /* Autostart PRG from D64 image */
          if (((D1551*)(machine->drive))->disk.image == NULL)
            {
              Message ( &gpDraw[0], "Attach a D64 image first !", 0x82, 0x80 );
            }
          else
            {
              int sel;
              char fname[32];

              if ((sel=SelectD64File ( machine, fname ))==-1)
                {
                  Message ( &gpDraw[0], "Loading cancelled", 0x82, 0x80 );
                } else
                {
                  int j;
                  u16 l = 0;

                  machine->RAM[0x0527+l++] = 76;  /* L */
                  machine->RAM[0x0527+l++] = 79;  /* O */
                  machine->RAM[0x0527+l++] = 65;  /* A */
                  machine->RAM[0x0527+l++] = 68;  /* D */
                  machine->RAM[0x0527+l++] = 34;  /* " */
                  for (j=0;j<strlen(fname);j++)
                    machine->RAM[0x0527+l++] = fname[j];
                  machine->RAM[0x0527+l++] = 34;  /* " */
                  machine->RAM[0x0527+l++] = 44;  /* , */
                  machine->RAM[0x0527+l++] = 56;  /* 8 */
                  machine->RAM[0x0527+l++] = 13;
                  machine->RAM[0x0527+l++] = 82;  /* R */
                  machine->RAM[0x0527+l++] = 85;  /* U */
                  machine->RAM[0x0527+l++] = 78;  /* N */
                  machine->RAM[0x0527+l++] = 13;
                  machine->RAM[0x00ef] = l;
                  return 1;
                }
            }
          break;
        case 3: /* Load PRG */
          return (LoadPRG ( machine ));
          break;
        case 4: /* Autostart PRG */
          {
            int r = LoadPRG ( machine );
            if (r==1)
              {
                P4_StartPRG ( machine );
                return 1;
              }
          }
          break;
        case 5: /* Screen brightness */
          if (k&GPC_VK_LEFT)
            {
              if (brightness>0)
                brightness--;
            } else
          if (k&GPC_VK_RIGHT)
            {
              if (brightness<16)
                brightness++;
            }
          sprintf(menu->pItems[5].name,"Screen brightness: +%d",brightness);
          SetPlus4Palette ( 16+brightness );
          break;
        case 7: /* Continue emulation */
          return 1;
          break;
      }
    return 0;
  }

void DoKb ( P4 * machine )
  {
    int k = GpKeyGet();
    BOOL b;

    b = (k&GPC_VK_FL)==GPC_VK_FL;
    if (k&GPC_VK_START)
      {
        GPMENU menu;
        GPMENUITEM items[8];
        char str[64];

        GpSoundBufStop();

        items[0].name = "Attach D64 image";
        items[0].color = 0x80;
        items[0].keys = GPC_VK_FA;
        items[1].name = "Load PRG from D64 image";
        items[1].color = 0x80;
        items[1].keys = GPC_VK_FA;
        items[2].name = "Autostart PRG from D64 image";
        items[2].color = 0x80;
        items[2].keys = GPC_VK_FA;
        items[3].name = "Load PRG";
        items[3].color = 0x80;
        items[3].keys = GPC_VK_FA;
        items[4].name = "Autostart PRG";
        items[4].color = 0x80;
        items[4].keys = GPC_VK_FA;
        sprintf(str,"Screen brightness: +%d",brightness);
        items[5].name = str;
        items[5].color = 0x80;
        items[5].keys = GPC_VK_LEFT|GPC_VK_RIGHT;
        items[6].name = NULL;
        items[7].name = "Continue emulation";
        items[7].color = 0x80;
        items[7].keys = GPC_VK_FA;
        menu.selection = menuselection;
        menu.pItems = items;
        menu.numItems = 8;
        menu.gpDraw = &gpDraw[0];
        menu.bgcolor = 0x82;
        menu.itemselcolor = 0x81;
        menu.xpos = 50;
        menu.ypos = 52;
        menu.userdata = (u32)machine;
        GpMenuStart ( &menu, menucallback );
        if (menu.selection == 7)
          menuselection = 0;
        else
          menuselection = menu.selection;
        InitKb(machine);
        GpSoundBufStart(&sb);
      } else
    if ( (k&GPC_VK_FL) && (k&GPC_VK_FR) ) /* Reset */
      {
        P4_Reset ( machine );
        while (GpKeyGet()&(GPC_VK_FL|GPC_VK_FR));
      } else
    if (!prevL && b)
      {
        mode++;
        if (mode>3) mode = 0;
        PaintArrow();
      }
    prevL = b;

    if (mode==0)  /* Keyboard */
      {
        if (k&(GPC_VK_UP|GPC_VK_DOWN|GPC_VK_LEFT|GPC_VK_RIGHT))
          {
            if (!prevM)
              {
                int lkbx = kbx,
                    lkby = kby;

                if (k&GPC_VK_RIGHT)
                  kbx++;
                else if (k&GPC_VK_LEFT)
                  kbx--;
                else if (k&GPC_VK_DOWN)
                  kby++;
                else if (k&GPC_VK_UP)
                      kby--;

                if (kbx>17)
                  kbx = 0;
                if (kbx<0)
                  kbx = 17;
                if (kby>4)
                  kby=0;
                if (kby<0)
                  kby=4;

                if ( (kbx!=lkbx) || (kby!=lkby) )
                  {
                    SetKbVal(machine,lkbx,lkby,0);
                    Inverse(lkbx,lkby);
                    Inverse(kbx,kby);
                  }
              }
            prevM = TRUE;
          } else
          {
            prevM = FALSE;
          }

          /* Return */
          machine->keys[0][1] = (k&GPC_VK_SELECT ? 1 : 0);
          /* Shift */
          machine->keys[1][7] = (k&GPC_VK_FR     ? 1 : 0);
          /* DEL */
          machine->keys[0][0] = (k&GPC_VK_FB     ? 1 : 0);

          if (k&GPC_VK_FA)
            SetKbVal(machine,kbx,kby,1);
          else
            SetKbVal(machine,kbx,kby,0);
        } else
      if (mode==1)  /* Joystick 1 */
        {
          machine->joy1[0] = (k&GPC_VK_UP    ? 1 : 0);
          machine->joy1[1] = (k&GPC_VK_DOWN  ? 1 : 0);
          machine->joy1[2] = (k&GPC_VK_LEFT  ? 1 : 0);
          machine->joy1[3] = (k&GPC_VK_RIGHT ? 1 : 0);
          machine->joy1[4] = (k&GPC_VK_FA    ? 1 : 0);
        } else
      if (mode==2)  /* Joystick 2 */
        {
          machine->joy2[0] = (k&GPC_VK_UP    ? 1 : 0);
          machine->joy2[1] = (k&GPC_VK_DOWN  ? 1 : 0);
          machine->joy2[2] = (k&GPC_VK_LEFT  ? 1 : 0);
          machine->joy2[3] = (k&GPC_VK_RIGHT ? 1 : 0);
          machine->joy2[4] = (k&GPC_VK_FA    ? 1 : 0);
        } else
      if (mode==3)  /* Cursor */
        {
          machine->keys[5][3] = (k&GPC_VK_UP    ? 1 : 0);
          machine->keys[5][0] = (k&GPC_VK_DOWN  ? 1 : 0);
          machine->keys[6][0] = (k&GPC_VK_LEFT  ? 1 : 0);
          machine->keys[6][3] = (k&GPC_VK_RIGHT ? 1 : 0);
        }
  }
