 /*
  * M6502.c - 6502 CPU core
  * Copyright (c)2002 Christian Nowak <chnowak@web.de>
  *
  *
  * This file is part of PlusIV.
  *
  * PlusIV is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * PlusIV is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with Foobar; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */

#include <stdio.h>
#include "M6502.h"
#include "M6502_NZ.h"
#include "cycles6502.h"
#include "defines.h"
/*
extern BOOL debug;
*/
#define READM(c,a)    c->readMem(c,a)
#define WRITEM(c,a,b) c->writeMem(c,a,b)

#define BYTE2WORD(a)  ((u16)(s16)(s8)a)

#define PUSH(c,a)     WRITEM(c,0x100|(u16)c->SP--,a)
#define POP(c)        READM(c,0x100|(u16)++c->SP)

/* #$ab */
#define IMMEDIATE(c)                (c->PC.aword+1)
/* $abcd */
#define ABSOLUTE(c)                 (READM(c,c->PC.aword+1) | (READM(c,c->PC.aword+2)<<8))
/* $ab */
#define ZP_ABSOLUTE(c)              ((u16)READM(c,c->PC.aword+1))
/* $abcd,xy */
#define INDEXED(c,xy)               (ABSOLUTE(c)+((u16)c->xy))
/* $ab,xy */
#define ZP_INDEXED(c,xy)            ((ZP_ABSOLUTE(c)+((u16)c->xy))&0xff)
/* ($ab,xy) */
#define PRE_INDEXED_INDIRECT(c,xy)  (READM(c,ZP_INDEXED(c,xy)) | (READM(c,ZP_INDEXED(c,xy+1))<<8))
/* ($ab),xy */
#define POST_INDEXED_INDIRECT(c,xy) ((READM(c,ZP_ABSOLUTE(c))|(READM(c,(ZP_ABSOLUTE(c)+1)&0xff)<<8)) + (u16)c->xy)

#define M_JR(c)   c->PC.aword += BYTE2WORD(READM(c,c->PC.aword+1))+2
#define UPDATE_NZ(c,v)\
  c->F = (c->F&~(F_N|F_Z)) | Table_NZ[v];

#define M_ADC(c,m)\
  {\
    u8 AH;\
    u8 AL;\
    tu16 = c->A+m+(c->F&F_C);\
    if (c->F&F_D)\
      {\
        AL=(c->A&0x0f)+(m&0x0f)+(c->F&F_C);\
        AH=(c->A>>4)+(m>>4);\
        if (AL>9)\
          {\
            AL+=6;\
            AH++;\
          }\
        c->F = (c->F&~(F_Z|F_N|F_V)) | ((tu16&0xff)?0:F_Z) | ((AH&0x08)?F_N:0) |\
          (((((AH<<4)^c->A)&0x80) && !((c->A^m)&0x80)) ? F_V : 0);\
        if (AH>9)\
          AH+=6;\
        c->F = (c->F&~F_C) | ((AH>0x0f)?F_C:0);\
        c->A=(AH<<4)|(AL&0x0f);\
      } else\
      {\
        c->F = (c->F&~(F_V|F_C)) | ((!((c->A^m)&0x80) && ((c->A^tu16)&0x80)) ? F_V:0 ) |\
          ((tu16>=0x100)?F_C:0);\
        c->A=tu16&0xff;\
        UPDATE_NZ(c,c->A);\
      }\
  }

#define M_SBC(c,m)\
  {\
    u16 dec_ac,bin_ac;\
    bin_ac=c->A-m-(1-(c->F&F_C));\
    if (c->F&F_D)\
      {\
        dec_ac=(c->A&0x0f)-(m&0x0f)-(1-(c->F&F_C));\
        if (dec_ac&0x10)\
          dec_ac=((dec_ac-6)&0x0f) | ((c->A&0xf0)-(m&0xf0)-0x10);\
        else\
          dec_ac=(dec_ac&0x0f) | ((c->A&0xf0)-(m&0xf0));\
        if (dec_ac&0x100)\
          dec_ac-=0x60;\
        c->F = (c->F&~(F_N|F_Z|F_V|F_C))|\
            ((bin_ac<0x100)?F_C:0) |\
            Table_NZ[(bin_ac&0xff)] |\
            ((((c->A^bin_ac)&0x80) && ((c->A^m)&0x80)) ? F_V:0 );\
        c->A=dec_ac&0xff;\
      }\
    else\
      {\
        c->F = (c->F&~(F_C|F_N|F_Z|F_V)) |\
          ((bin_ac<0x100)?F_C:0) |\
          ((((c->A^m)&0x80) && ((c->A^bin_ac)&0x80)) ? F_V:0 ) |\
          Table_NZ[(bin_ac&0xff)];\
        c->A=bin_ac&0xff;\
      }\
  }

#define M_AND(c,m)\
  c->A &= m;\
  c->F = (c->F & ~(F_N|F_Z)) | Table_NZ[c->A];

#define M_ASL(c,m)\
  tu8 = m << 1;\
  c->F = (c->F & ~(F_N|F_Z|F_C)) |\
          ((m&0x80)>>7) | Table_NZ[tu8];\
  m = tu8;

#define M_LSR(c,m)\
  tu8 = m >> 1;\
  c->F = (c->F & ~(F_N|F_Z|F_C)) |\
          (m&1) |\
Table_NZ[tu8];\
  m = tu8;

#define M_BIT(c,m)\
  tu8 = c->A & m;\
  c->F = (c->F & ~(F_N|F_Z|F_V)) |\
          (m&0x80) |\
          (m&0x40) |\
          (tu8==0?F_Z:0);

#define M_CMP(c,m,n)\
  tu16 = m-n;\
  c->F = (c->F & ~(F_N|F_Z|F_C)) |\
          (tu16<0x0100?F_C:0) |\
          Table_NZ[tu16&0xff];

#define M_DEC(c,m)\
  m--;\
  c->F = (c->F & ~(F_N|F_Z)) |\
  Table_NZ[m];

#define M_INC(c,m)\
  m++;\
  c->F = (c->F & ~(F_N|F_Z)) |\
    Table_NZ[m];

#define M_XOR(c,m)\
  c->A ^= m;\
  c->F = (c->F & ~(F_N|F_Z)) |\
  Table_NZ[c->A];

#define M_OR(c,m)\
  c->A |= m;\
  c->F = (c->F & ~(F_N|F_Z)) |\
  Table_NZ[c->A];

#define M_LD(c,r,m)\
  c->F = (c->F & ~(F_N|F_Z)) |\
  Table_NZ[c->r = m];

#define M_ROL(c,m)\
  tu8 =  m << 1;\
  tu8 |= c->F&F_C?1:0;\
  c->F = (c->F & ~(F_N|F_Z|F_C)) |\
        (m&0x80  ?F_C:0) |\
        Table_NZ[tu8];\
  m = tu8;

#define M_ROR(c,m)\
  tu8 = m >> 1;\
  tu8 |= (c->F&F_C)<<7;\
  c->F = (c->F & ~(F_N|F_Z|F_C)) |\
        (m&1) |\
        Table_NZ[tu8];\
  m = tu8;

/* INS (ISC) */
#define M_INS(c,m)\
  M_INC(c,m)\
  M_SBC(c,m)

/* DCM */
#define M_DCM(c,m)\
  m--;\
  M_CMP(c,c->A,m);


    /* For the macros */
    u8 tu8 = 0;
    u16 tu16 = 0;

    /* For the function */
    u8 mu8 = 0;
    u16 mu16 = 0;

#include "Op6502.h"

void M6502_Reset ( M6502 * cpu )
  {
    cpu->RES = FALSE;
    cpu->PC.abyte.low  = cpu->readMem(cpu,0xfffc);
    cpu->PC.abyte.high = cpu->readMem(cpu,0xfffd);
    cpu->A = 0;
    cpu->X = 0;
    cpu->Y = 0;
    cpu->F = 0x20 | F_I;
    cpu->SP = 0xff;
    cpu->IRQB = FALSE;
    cpu->NMIB = FALSE;
  }

void M6502_Init ( M6502 * cpu, u8 (*readMem)(M6502*,u16), void (*writeMem)(M6502*,u16,u8) )
  {
    cpu->PC.aword = 0;
    cpu->A = 0;
    cpu->X = 0;
    cpu->Y = 0;
    cpu->SP = 0;
    cpu->readMem =  (u8  (*)(void*,u16))    readMem;
    cpu->writeMem = (void(*)(void*,u16,u8)) writeMem;
    cpu->IRQB = FALSE;
    cpu->NMIB = FALSE;
    cpu->RES = TRUE;
    M6502_Reset(cpu);
  }


int M6502_Exec ( M6502 * cpu )
  {
    /* Handle NMIB */   /* NMI not needed for the Plus/4 */
/*    if (cpu->NMIB)
      {
        cpu->NMIB = FALSE;
        PUSH(cpu,cpu->PC.byte.high);
        PUSH(cpu,cpu->PC.byte.low);
        PUSH(cpu,cpu->F);
        cpu->F |= F_I;
        cpu->PC.byte.low  = READM(cpu, 0xfffa);
        cpu->PC.byte.high = READM(cpu, 0xfffb);
        return 7;
      }*/

    /* Handle IRQB */
/*    if (cpu->IRQB)
      {
        if (!(cpu->F&F_I))
          {
            cpu->IRQB = FALSE;
            PUSH(cpu,cpu->PC.byte.high);
            PUSH(cpu,cpu->PC.byte.low);
            PUSH(cpu,cpu->F&~F_B);
            cpu->F |= F_I;
            cpu->PC.byte.low  = READM(cpu, 0xfffe);
            cpu->PC.byte.high = READM(cpu, 0xffff);
            cpu->cycles=7;
          }
      }*/

    u8 instr;
#include "Exec6502.h"
    return cpu->cycles=cycles6502[instr];
  }

int M6502_Run ( M6502 * cpu, int c )
  {
    /* Handle IRQB (not cycle based) */
    if (cpu->IRQB)
      {
        if (!(cpu->F&F_I))
          {
            cpu->IRQB = FALSE;
            PUSH(cpu,cpu->PC.abyte.high);
            PUSH(cpu,cpu->PC.abyte.low);
            PUSH(cpu,cpu->F&~F_B);
            cpu->F |= F_I;
            cpu->PC.abyte.low  = READM(cpu, 0xfffe);
            cpu->PC.abyte.high = READM(cpu, 0xffff);
            cpu->cycles=7;
          }
      }
    cpu->cycles = 0;
    do
      {
        u8 instr;
/*        if (debug)
          printf("PC=%04x A=%02x X=%02x Y=%02x SP=%02x F=%02x\n", cpu->PC.aword, cpu->A, cpu->X, cpu->Y, cpu->SP, cpu->F);*/
#include "Exec6502.h"
        cpu->cycles += cycles6502[instr];
      } while (cpu->cycles<c);

    return cpu->cycles;
  }
