 /*
  * M6502.h - 6502 CPU core function and struct declarations
  * Copyright (c)2002 Christian Nowak <chnowak@web.de>
  *
  *
  * This file is part of PlusIV.
  *
  * PlusIV is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * PlusIV is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with Foobar; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */


#ifndef __M6502_H__
#define __M6502_H__

#include "defines.h"

#define F_N   0x80
#define F_V   0x40
#define F_B   0x10
#define F_D   0x08
#define F_I   0x04
#define F_Z   0x02
#define F_C   0x01

typedef struct M6502
  {
    union_word PC;  /* Program Counter */
    u8 A;     /* Accumulator */
    u8 X,Y;   /* Index registers */
    u8 SP;    /* Stack Pointer */
    u8 F;     /* Flags */
    u8   (*readMem) (void*,u16);
    void (*writeMem)(void*,u16,u8);
    BOOL IRQB;  /* IRQ */
    BOOL NMIB;  /* NMI */
    BOOL RES;   /* Reset */
    void * val;
    int cycles;
  } M6502;

void M6502_Init ( M6502 *, u8 (*)(M6502*,u16), void (*)(M6502*,u16,u8) );
void M6502_Reset ( M6502 * );
int M6502_Exec ( M6502 * );
int M6502_Run ( M6502 *, int );
#endif
