 /*
  * Op6502.h - 6502 CPU core operation codes implementations
  * Copyright (c)2002 Christian Nowak <chnowak@web.de>
  *
  *
  * This file is part of PlusIV.
  *
  * PlusIV is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * PlusIV is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with Foobar; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */

void OpIl(M6502 * cpu) {
/*    printf("Illegal opcode at %04x\n", cpu->PC.aword);*/
    cpu->PC.aword++;
  }

void Op00(M6502 * cpu) {  /* BRK */
          cpu->F|=F_B;
          cpu->PC.aword+=2;
          PUSH(cpu,cpu->PC.abyte.high);
          PUSH(cpu,cpu->PC.abyte.low);
          PUSH(cpu,cpu->F);
          cpu->F|=F_I;
          cpu->PC.abyte.low  = READM(cpu,0xfffe);
          cpu->PC.abyte.high = READM(cpu,0xffff);
  }

void Op01(M6502 * cpu) {
    M_OR(cpu,READM(cpu,PRE_INDEXED_INDIRECT(cpu,X))); cpu->PC.aword+=2;     /* ORA ($ab,X) */
  }

void Op04(M6502 * cpu) {
    cpu->PC.aword+=2;         /* SKB */
  }

void Op05(M6502 * cpu) {
    M_OR(cpu,READM(cpu,ZP_ABSOLUTE(cpu))); cpu->PC.aword+=2;              /* ORA $ab */
  }

void Op06(M6502 * cpu) {
    mu16=ZP_ABSOLUTE(cpu);mu8=READM(cpu,mu16);M_ASL(cpu,mu8);WRITEM(cpu,mu16,mu8);cpu->PC.aword+=2;       /* ASL $ab */
  }

void Op08(M6502 * cpu) {
    PUSH(cpu,cpu->F); cpu->PC.aword+=1;       /* PHP */
  }

void Op09(M6502 * cpu) {
    M_OR(cpu,READM(cpu,IMMEDIATE(cpu))); cpu->PC.aword+=2;                /* ORA #$ab */
  }

void Op0a(M6502 * cpu) {
    M_ASL(cpu,cpu->A);cpu->PC.aword+=1;                                           /* ASL A */
  }

void Op0c(M6502 * cpu) {
    cpu->PC.aword+=3;     /* SKW */
  }

void Op0d(M6502 * cpu) {
    M_OR(cpu,READM(cpu,ABSOLUTE(cpu))); cpu->PC.aword+=3;                   /* ORA $abcd */
  }

void Op0e(M6502 * cpu) {
    mu16=ABSOLUTE(cpu);mu8=READM(cpu,mu16);M_ASL(cpu,mu8);WRITEM(cpu,mu16,mu8);cpu->PC.aword+=3;        /* ASL $abcd */
  }

void Op10(M6502 * cpu) {
    if (cpu->F&F_N) cpu->PC.aword+=2; else M_JR(cpu);   /* BPL */
  }

void Op11(M6502 * cpu) {
    M_OR(cpu,READM(cpu,POST_INDEXED_INDIRECT(cpu,Y))); cpu->PC.aword+=2;  /* ORA ($ab),Y */
  }

void Op14(M6502 * cpu) {
    cpu->PC.aword+=2;         /* SKB */
  }

void Op15(M6502 * cpu) {
    M_OR(cpu,READM(cpu,ZP_INDEXED(cpu,X))); cpu->PC.aword+=2;               /* ORA $ab,X */
  }

void Op16(M6502 * cpu) {
    mu16=ZP_INDEXED(cpu,X);mu8=READM(cpu,mu16);M_ASL(cpu,mu8);WRITEM(cpu,mu16,mu8);cpu->PC.aword+=2;    /* ASL $ab,X */
  }

void Op18(M6502 * cpu) {
    cpu->F &= ~F_C; cpu->PC.aword+=1;                   /* CLC */
  }

void Op19(M6502 * cpu) {
    M_OR(cpu,READM(cpu,INDEXED(cpu,Y))); cpu->PC.aword+=3;                /* ORA $abcd,Y */
  }

void Op1a(M6502 * cpu) {
    cpu->PC.aword+=1;         /* NOP */
  }

void Op1c(M6502 * cpu) {
    cpu->PC.aword+=3;     /* SKW */
  }

void Op1d(M6502 * cpu) {
    M_OR(cpu,READM(cpu,INDEXED(cpu,X))); cpu->PC.aword+=3;                /* ORA $abcd,X */
  }

void Op1e(M6502 * cpu) {
    mu16=INDEXED(cpu,X);mu8=READM(cpu,mu16);M_ASL(cpu,mu8);WRITEM(cpu,mu16,mu8);cpu->PC.aword+=3;         /* ASL $abcd,X */
  }

void Op20(M6502 * cpu) {
    mu16=ABSOLUTE(cpu); cpu->PC.aword+=2; PUSH(cpu,cpu->PC.abyte.high); PUSH(cpu,cpu->PC.abyte.low); cpu->PC.aword = mu16;    /* JSR $abcd */
  }

void Op21(M6502 * cpu) {
    M_AND(cpu,READM(cpu,PRE_INDEXED_INDIRECT(cpu,X)));cpu->PC.aword+=2;     /* AND ($ab,X) */
  }

void Op24(M6502 * cpu) {
    mu8=READM(cpu,ZP_ABSOLUTE(cpu));M_BIT(cpu,mu8);cpu->PC.aword+=2;    /* BIT $ab */
  }

void Op25(M6502 * cpu) {
    M_AND(cpu,READM(cpu,ZP_ABSOLUTE(cpu))); cpu->PC.aword+=2;             /* AND $ab */
  }

void Op26(M6502 * cpu) {
    mu16=ZP_ABSOLUTE(cpu);mu8=READM(cpu,mu16);M_ROL(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=2;        /* ROL $ab */
  }

void Op28(M6502 * cpu) {
    cpu->F=POP(cpu)|0x20; cpu->PC.aword+=1;         /* PLP */
  }

void Op29(M6502 * cpu) {
    M_AND(cpu,READM(cpu,IMMEDIATE(cpu))); cpu->PC.aword+=2;               /* AND #$ab */
  }

void Op2a(M6502 * cpu) {
    M_ROL(cpu,cpu->A); cpu->PC.aword+=1;                                                                                  /* ROL A */
  }

void Op2c(M6502 * cpu) {
    mu8=READM(cpu,ABSOLUTE(cpu));M_BIT(cpu,mu8);cpu->PC.aword+=3;       /* BIT $abcd */
  }

void Op2d(M6502 * cpu) {
    M_AND(cpu,READM(cpu,ABSOLUTE(cpu))); cpu->PC.aword+=3;              /* AND $abcd */
  }

void Op2e(M6502 * cpu) {
    mu16=ABSOLUTE(cpu);mu8=READM(cpu,mu16);M_ROL(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=3;           /* ROL $abcd */
  }

void Op30(M6502 * cpu) {
    if (cpu->F&F_N) M_JR(cpu); else cpu->PC.aword+=2;   /* BMI */
  }

void Op31(M6502 * cpu) {
    M_AND(cpu,READM(cpu,POST_INDEXED_INDIRECT(cpu,Y)));cpu->PC.aword+=2;  /* AND ($ab),Y */
  }

void Op34(M6502 * cpu) {
    cpu->PC.aword+=2;         /* SKB */
  }

void Op35(M6502 * cpu) {
    M_AND(cpu,READM(cpu,ZP_INDEXED(cpu,X))); cpu->PC.aword+=2;          /* AND $ab,X */
  }

void Op36(M6502 * cpu) {
    mu16=ZP_INDEXED(cpu,X);mu8=READM(cpu,mu16);M_ROL(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=2;       /* ROL $ab,X */
  }

void Op38(M6502 * cpu) {
    cpu->F |= F_C; cpu->PC.aword+=1;    /* SEC */
  }

void Op39(M6502 * cpu) {
    M_AND(cpu,READM(cpu,INDEXED(cpu,Y))); cpu->PC.aword+=3;               /* AND $abcd,Y */
  }

void Op3a(M6502 * cpu) {
    cpu->PC.aword+=1;         /* NOP */
  }

void Op3c(M6502 * cpu) {
    cpu->PC.aword+=3;     /* SKW */
  }

void Op3d(M6502 * cpu) {
    M_AND(cpu,READM(cpu,INDEXED(cpu,X))); cpu->PC.aword+=3;               /* AND $abcd,X */
  }

void Op3e(M6502 * cpu) {
    mu16=INDEXED(cpu,X);mu8=READM(cpu,mu16);M_ROL(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=3;          /* ROL $abcd,X */
  }

void Op40(M6502 * cpu) {
    cpu->F=POP(cpu);cpu->PC.abyte.low=POP(cpu);cpu->PC.abyte.high=POP(cpu);     /* RTI */
  }

void Op41(M6502 * cpu) {
    M_XOR(cpu,READM(cpu,PRE_INDEXED_INDIRECT(cpu,X))); cpu->PC.aword+=2;  /* EOR ($ab,X) */
  }

void Op44(M6502 * cpu) {
    cpu->PC.aword+=2;         /* SKB */
  }

void Op45(M6502 * cpu) {
    M_XOR(cpu,READM(cpu,ZP_ABSOLUTE(cpu))); cpu->PC.aword+=2;               /* EOR $ab */
  }

void Op46(M6502 * cpu) {
    mu16=ZP_ABSOLUTE(cpu);mu8=READM(cpu,mu16);M_LSR(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=2;        /* LSR $ab */
  }

void Op48(M6502 * cpu) {
    PUSH(cpu,cpu->A); cpu->PC.aword+=1;       /* PHA */
  }

void Op49(M6502 * cpu) {
    M_XOR(cpu,READM(cpu,IMMEDIATE(cpu))); cpu->PC.aword+=2;                 /* EOR #$ab */
  }

void Op4a(M6502 * cpu) {
    M_LSR(cpu,cpu->A); cpu->PC.aword+=1;        /* LSR A */
  }

void Op4c(M6502 * cpu) {
    mu16=ABSOLUTE(cpu); cpu->PC.aword=mu16;                                                                   /* JMP $abcd */
  }

void Op4d(M6502 * cpu) {
    M_XOR(cpu,READM(cpu,ABSOLUTE(cpu))); cpu->PC.aword+=3;                /* EOR $abcd */
  }

void Op4e(M6502 * cpu) {
    mu16=ABSOLUTE(cpu);mu8=READM(cpu,mu16);M_LSR(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=3;           /* LSR $abcd */
  }

void Op50(M6502 * cpu) {
    if (cpu->F&F_V) cpu->PC.aword+=2; else M_JR(cpu);   /* BVC */
  }

void Op51(M6502 * cpu) {
    M_XOR(cpu,READM(cpu,POST_INDEXED_INDIRECT(cpu,Y))); cpu->PC.aword+=2;   /* EOR ($ab),Y */
  }

void Op54(M6502 * cpu) {
    cpu->PC.aword+=2;         /* SKB */
  }

void Op55(M6502 * cpu) {
    M_XOR(cpu,READM(cpu,ZP_INDEXED(cpu,X))); cpu->PC.aword+=2;            /* EOR $ab,X */
  }

void Op56(M6502 * cpu) {
    mu16=ZP_INDEXED(cpu,X);mu8=READM(cpu,mu16);M_LSR(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=2;       /* LSR $ab,X */
  }

void Op58(M6502 * cpu) {
    cpu->F &= ~F_I; cpu->PC.aword+=1;                   /* CLI */
  }

void Op59(M6502 * cpu) {
    M_XOR(cpu,READM(cpu,INDEXED(cpu,Y))); cpu->PC.aword+=3;                 /* EOR $abcd,Y */
  }

void Op5a(M6502 * cpu) {
    cpu->PC.aword+=1;         /* NOP */
  }

void Op5c(M6502 * cpu) {
    cpu->PC.aword+=3;     /* SKW */
  }

void Op5d(M6502 * cpu) {
    M_XOR(cpu,READM(cpu,INDEXED(cpu,X))); cpu->PC.aword+=3;                 /* EOR $abcd,X */
  }

void Op5e(M6502 * cpu) {
    mu16=INDEXED(cpu,X);mu8=READM(cpu,mu16);M_LSR(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=3;          /* LSR $abcd,X */
  }

void Op60(M6502 * cpu) {
    cpu->PC.abyte.low=POP(cpu);cpu->PC.abyte.high=POP(cpu); cpu->PC.aword+=1;                     /* RTS */
  }

void Op61(M6502 * cpu) {
    mu8=READM(cpu,PRE_INDEXED_INDIRECT(cpu,X));M_ADC(cpu,mu8); cpu->PC.aword+=2;  /* ADC ($ab,X) */
  }

void Op64(M6502 * cpu) {
    cpu->PC.aword+=2;         /* SKB */
  }

void Op65(M6502 * cpu) {
    mu8=READM(cpu,ZP_ABSOLUTE(cpu));M_ADC(cpu,mu8); cpu->PC.aword+=2;               /* ADC $ab */
  }

void Op66(M6502 * cpu) {
    mu16=ZP_ABSOLUTE(cpu);mu8=READM(cpu,mu16);M_ROR(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=2;        /* ROR $ab */
  }

void Op68(M6502 * cpu) {
    cpu->A=POP(cpu); cpu->F=(cpu->F&~(F_N|F_Z))|Table_NZ[cpu->A]; cpu->PC.aword+=1;         /* PLA */
  }

void Op69(M6502 * cpu) {
    mu8=READM(cpu,IMMEDIATE(cpu));M_ADC(cpu,mu8); cpu->PC.aword+=2;                 /* ADC #$ab */
  }

void Op6a(M6502 * cpu) {
    M_ROR(cpu,cpu->A); cpu->PC.aword+=1;                                                                                  /* ROR A */
  }

void Op6c(M6502 * cpu) {
    mu16=ABSOLUTE(cpu); cpu->PC.abyte.low=READM(cpu,mu16);cpu->PC.abyte.high=READM(cpu,mu16+1);   /* JMP ($abcd) */
  }

void Op6d(M6502 * cpu) {
    mu8=READM(cpu,ABSOLUTE(cpu));M_ADC(cpu,mu8); cpu->PC.aword+=3;                /* ADC $abcd */
  }

void Op6e(M6502 * cpu) {
    mu16=ABSOLUTE(cpu);mu8=READM(cpu,mu16);M_ROR(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=3;           /* ROR $abcd */
  }

void Op70(M6502 * cpu) {
    if (cpu->F&F_V) M_JR(cpu); else cpu->PC.aword+=2;   /* BVS */
  }

void Op71(M6502 * cpu) {
    mu8=READM(cpu,POST_INDEXED_INDIRECT(cpu,Y));M_ADC(cpu,mu8); cpu->PC.aword+=2;   /* ADC ($ab),Y */
  }

void Op74(M6502 * cpu) {
    cpu->PC.aword+=2;         /* SKB */
  }

void Op75(M6502 * cpu) {
    mu8=READM(cpu,ZP_INDEXED(cpu,X));M_ADC(cpu,mu8); cpu->PC.aword+=2;            /* ADC $ab,X */
  }

void Op76(M6502 * cpu) {
    mu16=ZP_INDEXED(cpu,X);mu8=READM(cpu,mu16);M_ROR(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=2;       /* ROR $ab,X */
  }

void Op78(M6502 * cpu) {
    cpu->F |= F_I; cpu->PC.aword+=1;    /* SEI */
  }

void Op79(M6502 * cpu) {
    mu8=READM(cpu,INDEXED(cpu,Y));M_ADC(cpu,mu8); cpu->PC.aword+=3;                 /* ADC $abcd,Y */
  }

void Op7a(M6502 * cpu) {
    cpu->PC.aword+=1;         /* NOP */
  }

void Op7c(M6502 * cpu) {
    cpu->PC.aword+=3;     /* SKW */
  }

void Op7d(M6502 * cpu) {
    mu8=READM(cpu,INDEXED(cpu,X));M_ADC(cpu,mu8); cpu->PC.aword+=3;                 /* ADC $abcd,X */
  }

void Op7e(M6502 * cpu) {
    mu16=INDEXED(cpu,X);mu8=READM(cpu,mu16);M_ROR(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=3;          /* ROR $abcd,X */
  }

void Op80(M6502 * cpu) {
    cpu->PC.aword+=2;         /* SKB */
  }

void Op81(M6502 * cpu) {
    WRITEM(cpu,PRE_INDEXED_INDIRECT(cpu,X),cpu->A); cpu->PC.aword+=2;     /* STA ($ab,X) */
  }

void Op82(M6502 * cpu) {
    cpu->PC.aword+=2;         /* SKB */
  }

void Op83(M6502 * cpu) {
    WRITEM(cpu,PRE_INDEXED_INDIRECT(cpu,X),cpu->A&cpu->X); cpu->PC.aword+=2;  /* AXS ($ab,X) */
  }

void Op84(M6502 * cpu) {
    WRITEM(cpu,ZP_ABSOLUTE(cpu),cpu->Y); cpu->PC.aword+=2;              /* STY $ab */
  }

void Op85(M6502 * cpu) {
    WRITEM(cpu,ZP_ABSOLUTE(cpu),cpu->A); cpu->PC.aword+=2;              /* STA $ab */
  }

void Op86(M6502 * cpu) {
    WRITEM(cpu,ZP_ABSOLUTE(cpu),cpu->X); cpu->PC.aword+=2;              /* STX $ab */
  }

void Op87(M6502 * cpu) {
    WRITEM(cpu,ZP_ABSOLUTE(cpu),cpu->A&cpu->X); cpu->PC.aword+=2;       /* AXS $ab */
  }

void Op88(M6502 * cpu) {
    M_DEC(cpu,cpu->Y); cpu->PC.aword+=1;        /* DEY */
  }

void Op8a(M6502 * cpu) {
    cpu->A=cpu->X; UPDATE_NZ(cpu,cpu->A); cpu->PC.aword+=1;       /* TXA */
  }

void Op8c(M6502 * cpu) {
    WRITEM(cpu,ABSOLUTE(cpu),cpu->Y); cpu->PC.aword+=3;                   /* STY $abcd */
  }

void Op8d(M6502 * cpu) {
    WRITEM(cpu,ABSOLUTE(cpu),cpu->A); cpu->PC.aword+=3;                   /* STA $abcd */
  }

void Op8e(M6502 * cpu) {
    WRITEM(cpu,ABSOLUTE(cpu),cpu->X); cpu->PC.aword+=3;                   /* STX $abcd */
  }

void Op8f(M6502 * cpu) {
    WRITEM(cpu,ABSOLUTE(cpu),cpu->A&cpu->X); cpu->PC.aword+=3;            /* AXS $abcd */
  }

void Op90(M6502 * cpu) {
    if (cpu->F&F_C) cpu->PC.aword+=2; else M_JR(cpu);     /* BCC */
  }

void Op91(M6502 * cpu) {
    WRITEM(cpu,POST_INDEXED_INDIRECT(cpu,Y),cpu->A); cpu->PC.aword+=2;  /* STA ($ab),Y */
  }

void Op94(M6502 * cpu) {
    WRITEM(cpu,ZP_INDEXED(cpu,X),cpu->Y); cpu->PC.aword+=2;               /* STY $ab,X */
  }

void Op95(M6502 * cpu) {
    WRITEM(cpu,ZP_INDEXED(cpu,X),cpu->A); cpu->PC.aword+=2;               /* STA $ab,X */
  }

void Op96(M6502 * cpu) {
    WRITEM(cpu,ZP_INDEXED(cpu,Y),cpu->X); cpu->PC.aword+=2;               /* STX $ab,Y */
  }

void Op97(M6502 * cpu) {
    WRITEM(cpu,ZP_INDEXED(cpu,Y),cpu->A&cpu->X); cpu->PC.aword+=2;        /* AXS $ab,Y */
  }

void Op98(M6502 * cpu) {
    cpu->A=cpu->Y; UPDATE_NZ(cpu,cpu->A); cpu->PC.aword+=1;       /* TYA */
  }

void Op99(M6502 * cpu) {
    WRITEM(cpu,INDEXED(cpu,Y),cpu->A); cpu->PC.aword+=3;                /* STA $abcd,Y */
  }

void Op9a(M6502 * cpu) {
    cpu->SP=cpu->X;                   cpu->PC.aword+=1;       /* TXS */
  }

void Op9d(M6502 * cpu) {
    WRITEM(cpu,INDEXED(cpu,X),cpu->A); cpu->PC.aword+=3;                /* STA $abcd,X */
  }

void Opa0(M6502 * cpu) {
    M_LD(cpu,Y,READM(cpu,IMMEDIATE(cpu))); cpu->PC.aword+=2;                /* LDY #$ab */
  }

void Opa1(M6502 * cpu) {
    M_LD(cpu,A,READM(cpu,PRE_INDEXED_INDIRECT(cpu,X))); cpu->PC.aword+=2;     /* LDA ($ab,X) */
  }

void Opa2(M6502 * cpu) {
    M_LD(cpu,X,READM(cpu,IMMEDIATE(cpu))); cpu->PC.aword+=2;                /* LDX #$ab */
  }

void Opa3(M6502 * cpu) {
    M_LD(cpu,A,READM(cpu,PRE_INDEXED_INDIRECT(cpu,X)));cpu->X=cpu->A; cpu->PC.aword+=2;     /* LAX ($ab,X) */
  }

void Opa4(M6502 * cpu) {
    M_LD(cpu,Y,READM(cpu,ZP_ABSOLUTE(cpu))); cpu->PC.aword+=2;              /* LDY $ab */
  }

void Opa5(M6502 * cpu) {
    M_LD(cpu,A,READM(cpu,ZP_ABSOLUTE(cpu))); cpu->PC.aword+=2;                /* LDA $ab */
  }

void Opa6(M6502 * cpu) {
    M_LD(cpu,X,READM(cpu,ZP_ABSOLUTE(cpu))); cpu->PC.aword+=2;              /* LDX $ab */
  }

void Opa7(M6502 * cpu) {
    M_LD(cpu,A,READM(cpu,ZP_ABSOLUTE(cpu)));cpu->X=cpu->A; cpu->PC.aword+=2;              /* LAX $ab */
  }

void Opa8(M6502 * cpu) {
    cpu->Y=cpu->A; UPDATE_NZ(cpu,cpu->Y); cpu->PC.aword+=1;       /* TAY */
  }

void Opa9(M6502 * cpu) {
    M_LD(cpu,A,READM(cpu,IMMEDIATE(cpu))); cpu->PC.aword+=2;                /* LDA #$ab */
  }

void Opaa(M6502 * cpu) {
    cpu->X=cpu->A; UPDATE_NZ(cpu,cpu->X); cpu->PC.aword+=1;       /* TAX */
  }

void Opac(M6502 * cpu) {
    M_LD(cpu,Y,READM(cpu,ABSOLUTE(cpu))); cpu->PC.aword+=3;                   /* LDY $abcd */
  }

void Opad(M6502 * cpu) {
    M_LD(cpu,A,READM(cpu,ABSOLUTE(cpu))); cpu->PC.aword+=3;                   /* LDA $abcd */
  }

void Opae(M6502 * cpu) {
    M_LD(cpu,X,READM(cpu,ABSOLUTE(cpu))); cpu->PC.aword+=3;                   /* LDX $abcd */
  }

void Opaf(M6502 * cpu) {
    M_LD(cpu,A,READM(cpu,ABSOLUTE(cpu)));cpu->X=cpu->A; cpu->PC.aword+=3;                   /* LAX $abcd */
  }

void Opb0(M6502 * cpu) {
    if (cpu->F&F_C) M_JR(cpu); else cpu->PC.aword+=2;     /* BCS */
  }

void Opb1(M6502 * cpu) {
    M_LD(cpu,A,READM(cpu,POST_INDEXED_INDIRECT(cpu,Y))); cpu->PC.aword+=2;  /* LDA ($ab),Y */
  }

void Opb3(M6502 * cpu) {
    M_LD(cpu,A,READM(cpu,POST_INDEXED_INDIRECT(cpu,Y)));cpu->X=cpu->A; cpu->PC.aword+=2;  /* LAX ($ab),Y */
  }

void Opb4(M6502 * cpu) {
    M_LD(cpu,Y,READM(cpu,ZP_INDEXED(cpu,X))); cpu->PC.aword+=2;               /* LDY $ab,X */
  }

void Opb5(M6502 * cpu) {
    M_LD(cpu,A,READM(cpu,ZP_INDEXED(cpu,X))); cpu->PC.aword+=2;               /* LDA $ab,X */
  }

void Opb6(M6502 * cpu) {
    M_LD(cpu,X,READM(cpu,ZP_INDEXED(cpu,Y))); cpu->PC.aword+=2;               /* LDX $ab,Y */
  }

void Opb7(M6502 * cpu) {
    M_LD(cpu,A,READM(cpu,ZP_INDEXED(cpu,Y)));cpu->X=cpu->A; cpu->PC.aword+=2;               /* LAX $ab,Y */
  }

void Opb8(M6502 * cpu) {
    cpu->F &= ~F_V; cpu->PC.aword+=1;                   /* CLV */
  }

void Opb9(M6502 * cpu) {
    M_LD(cpu,A,READM(cpu,INDEXED(cpu,Y))); cpu->PC.aword+=3;                /* LDA $abcd,Y */
  }

void Opba(M6502 * cpu) {
    cpu->X=cpu->SP;UPDATE_NZ(cpu,cpu->X); cpu->PC.aword+=1;       /* TSX */
  }

void Opbc(M6502 * cpu) {
    M_LD(cpu,Y,READM(cpu,INDEXED(cpu,X))); cpu->PC.aword+=3;                /* LDY $abcd,X */
  }

void Opbd(M6502 * cpu) {
    M_LD(cpu,A,READM(cpu,INDEXED(cpu,X))); cpu->PC.aword+=3;                /* LDA $abcd,X */
  }

void Opbe(M6502 * cpu) {
    M_LD(cpu,X,READM(cpu,INDEXED(cpu,Y))); cpu->PC.aword+=3;                /* LDX $abcd,Y */
  }

void Opbf(M6502 * cpu) {
    M_LD(cpu,A,READM(cpu,INDEXED(cpu,Y)));cpu->X=cpu->A; cpu->PC.aword+=3;                /* LAX $abcd,Y */
  }

void Opc0(M6502 * cpu) {
    M_CMP(cpu,cpu->Y,READM(cpu,IMMEDIATE(cpu))); cpu->PC.aword+=2;                /* CPY #$ab */
  }

void Opc1(M6502 * cpu) {
    M_CMP(cpu,cpu->A,READM(cpu,PRE_INDEXED_INDIRECT(cpu,X))); cpu->PC.aword+=2;     /* CMP ($ab,X) */
  }

void Opc2(M6502 * cpu) {
    cpu->PC.aword+=2;         /* SKB */
  }

void Opc3(M6502 * cpu) {
    mu16=PRE_INDEXED_INDIRECT(cpu,X);mu8=READM(cpu,mu16);M_DCM(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=2; /* DCM ($ab,X) */
  }

void Opc4(M6502 * cpu) {
    M_CMP(cpu,cpu->Y,READM(cpu,ZP_ABSOLUTE(cpu))); cpu->PC.aword+=2;              /* CPY $ab */
  }

void Opc5(M6502 * cpu) {
    M_CMP(cpu,cpu->A,READM(cpu,ZP_ABSOLUTE(cpu))); cpu->PC.aword+=2;              /* CMP $ab */
  }

void Opc6(M6502 * cpu) {
    mu16=ZP_ABSOLUTE(cpu);mu8=READM(cpu,mu16);M_DEC(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=2;      /* DEC $ab */
  }

void Opc7(M6502 * cpu) {
    mu16=ZP_ABSOLUTE(cpu);mu8=READM(cpu,mu16);M_DCM(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=2;        /* DCM $ab */
  }

void Opc8(M6502 * cpu) {
    M_INC(cpu,cpu->Y); cpu->PC.aword+=1;        /* INY */
  }

void Opc9(M6502 * cpu) {
    M_CMP(cpu,cpu->A,READM(cpu,IMMEDIATE(cpu))); cpu->PC.aword+=2;                /* CMP #$ab */
  }

void Opca(M6502 * cpu) {
    M_DEC(cpu,cpu->X); cpu->PC.aword+=1;        /* DEX */
  }

void Opcc(M6502 * cpu) {
    M_CMP(cpu,cpu->Y,READM(cpu,ABSOLUTE(cpu))); cpu->PC.aword+=3;                   /* CPY $abcd */
  }

void Opcd(M6502 * cpu) {
    M_CMP(cpu,cpu->A,READM(cpu,ABSOLUTE(cpu))); cpu->PC.aword+=3;                   /* CMP $abcd */
  }

void Opce(M6502 * cpu) {
    mu16=ABSOLUTE(cpu);mu8=READM(cpu,mu16);M_DEC(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=3;         /* DEC $abcd */
  }

void Opcf(M6502 * cpu) {
    mu16=ABSOLUTE(cpu);mu8=READM(cpu,mu16);M_DCM(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=3;       /* DCM $abcd */
  }

void Opd0(M6502 * cpu) {
    if (cpu->F&F_Z) cpu->PC.aword+=2; else M_JR(cpu);   /* BNE */
  }

void Opd1(M6502 * cpu) {
    M_CMP(cpu,cpu->A,READM(cpu,POST_INDEXED_INDIRECT(cpu,Y))); cpu->PC.aword+=2;    /* CMP ($ab),Y */
  }

void Opd3(M6502 * cpu) {
    mu16=POST_INDEXED_INDIRECT(cpu,Y);mu8=READM(cpu,mu16);M_DCM(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=2;    /* DCM ($ab),Y */
  }

void Opd4(M6502 * cpu) {
    cpu->PC.aword+=2;         /* SKB */
  }

void Opd5(M6502 * cpu) {
    M_CMP(cpu,cpu->A,READM(cpu,ZP_INDEXED(cpu,X))); cpu->PC.aword+=2;               /* CMP $ab,X */
  }

void Opd6(M6502 * cpu) {
    mu16=ZP_INDEXED(cpu,X);mu8=READM(cpu,mu16);M_DEC(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=2;     /* DEC $ab,X */
  }

void Opd7(M6502 * cpu) {
    mu16=ZP_INDEXED(cpu,X);mu8=READM(cpu,mu16);M_DCM(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=2;   /* DCM $ab,X */
  }

void Opd8(M6502 * cpu) {
    cpu->F &= ~F_D; cpu->PC.aword+=1;                   /* CLD */
  }

void Opd9(M6502 * cpu) {
    M_CMP(cpu,cpu->A,READM(cpu,INDEXED(cpu,Y))); cpu->PC.aword+=3;                /* CMP $abcd,Y */
  }

void Opda(M6502 * cpu) {
    cpu->PC.aword+=1;         /* NOP */
  }

void Opdb(M6502 * cpu) {
    mu16=INDEXED(cpu,Y);mu8=READM(cpu,mu16);M_DCM(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=3;          /* DCM $abcd,Y */
  }

void Opdc(M6502 * cpu) {
    cpu->PC.aword+=3;     /* SKW */
  }

void Opdd(M6502 * cpu) {
    M_CMP(cpu,cpu->A,READM(cpu,INDEXED(cpu,X))); cpu->PC.aword+=3;                /* CMP $abcd,X */
  }

void Opde(M6502 * cpu) {
    mu16=INDEXED(cpu,X);mu8=READM(cpu,mu16);M_DEC(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=3;        /* DEC $abcd,X */
  }

void Opdf(M6502 * cpu) {
    mu16=INDEXED(cpu,X);mu8=READM(cpu,mu16);M_DCM(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=3;          /* DCM $abcd,X */
  }

void Ope0(M6502 * cpu) {
    M_CMP(cpu,cpu->X,READM(cpu,IMMEDIATE(cpu))); cpu->PC.aword+=2;                /* CPX #$ab */
  }

void Ope1(M6502 * cpu) {
    mu8=READM(cpu,PRE_INDEXED_INDIRECT(cpu,X));M_SBC(cpu,mu8); cpu->PC.aword+=2;  /* SBC ($ab,X) */
  }

void Ope2(M6502 * cpu) {
    cpu->PC.aword+=2;         /* SKB */
  }

void Ope3(M6502 * cpu) {
    mu16=PRE_INDEXED_INDIRECT(cpu,X);mu8=READM(cpu,mu16);M_INS(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=2;   /* INS ($ab,X) */
  }

void Ope4(M6502 * cpu) {
    M_CMP(cpu,cpu->X,READM(cpu,ZP_ABSOLUTE(cpu))); cpu->PC.aword+=2;              /* CPX $ab */
  }

void Ope5(M6502 * cpu) {
    mu8=READM(cpu,ZP_ABSOLUTE(cpu));M_SBC(cpu,mu8); cpu->PC.aword+=2;               /* SBC $ab */
  }

void Ope6(M6502 * cpu) {
    mu16=ZP_ABSOLUTE(cpu);mu8=READM(cpu,mu16);M_INC(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=2;      /* INC $ab */
  }

void Ope7(M6502 * cpu) {
    mu16=ZP_ABSOLUTE(cpu);mu8=READM(cpu,mu16);M_INS(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=2;              /* INS $ab */
  }

void Ope8(M6502 * cpu) {
    M_INC(cpu,cpu->X); cpu->PC.aword+=1;        /* INX */
  }

void Ope9(M6502 * cpu) {
    mu8=READM(cpu,IMMEDIATE(cpu));M_SBC(cpu,mu8); cpu->PC.aword+=2;                 /* SBC #$ab */
  }

void Opea(M6502 * cpu) {
    cpu->PC.aword+=1;     /* NOP */
  }

void Opec(M6502 * cpu) {
    M_CMP(cpu,cpu->X,READM(cpu,ABSOLUTE(cpu))); cpu->PC.aword+=3;                   /* CPX $abcd */
  }

void Oped(M6502 * cpu) {
    mu8=READM(cpu,ABSOLUTE(cpu));M_SBC(cpu,mu8); cpu->PC.aword+=3;                /* SBC $abcd */
  }

void Opee(M6502 * cpu) {
    mu16=ABSOLUTE(cpu);mu8=READM(cpu,mu16);M_INC(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=3;         /* INC $abcd */
  }

void Opef(M6502 * cpu) {
    mu16=ABSOLUTE(cpu);mu8=READM(cpu,mu16);M_INS(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=3;                 /* INS $abcd */
  }

void Opf0(M6502 * cpu) {
    if (cpu->F&F_Z) M_JR(cpu); else cpu->PC.aword+=2;     /* BEQ */
  }

void Opf1(M6502 * cpu) {
    mu8=READM(cpu,POST_INDEXED_INDIRECT(cpu,Y));M_SBC(cpu,mu8); cpu->PC.aword+=2;   /* SBC ($ab),Y */
  }

void Opf3(M6502 * cpu) {
    mu16=POST_INDEXED_INDIRECT(cpu,Y);mu8=READM(cpu,mu16);M_INS(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=2;  /* INS ($ab),Y */
  }

void Opf4(M6502 * cpu) {
    cpu->PC.aword+=2;         /* SKB */
  }

void Opf5(M6502 * cpu) {
    mu8=READM(cpu,ZP_INDEXED(cpu,X));M_SBC(cpu,mu8); cpu->PC.aword+=2;            /* SBC $ab,X */
  }

void Opf6(M6502 * cpu) {
    mu16=ZP_INDEXED(cpu,X);mu8=READM(cpu,mu16);M_INC(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=2;     /* INC $ab,X */
  }

void Opf7(M6502 * cpu) {
    mu16=ZP_INDEXED(cpu,X);mu8=READM(cpu,mu16);M_INS(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=2;             /* INS $ab,X */
  }

void Opf8(M6502 * cpu) {
    cpu->F |= F_D; cpu->PC.aword+=1;    /* SED */
  }

void Opf9(M6502 * cpu) {
    mu8=READM(cpu,INDEXED(cpu,Y));M_SBC(cpu,mu8); cpu->PC.aword+=3;                 /* SBC $abcd,Y */
  }

void Opfa(M6502 * cpu) {
    cpu->PC.aword+=1;         /* NOP */
  }

void Opfb(M6502 * cpu) {
    mu16=INDEXED(cpu,Y);mu8=READM(cpu,mu16);M_INS(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=3;                /* INS $abcd,Y */
  }

void Opfc(M6502 * cpu) {
    cpu->PC.aword+=3;     /* SKW */
  }

void Opfd(M6502 * cpu) {
    mu8=READM(cpu,INDEXED(cpu,X));M_SBC(cpu,mu8); cpu->PC.aword+=3;                 /* SBC $abcd,X */
  }

void Opfe(M6502 * cpu) {
    mu16=INDEXED(cpu,X);mu8=READM(cpu,mu16);M_INC(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=3;        /* INC $abcd,X */
  }

void Opff(M6502 * cpu) {
    mu16=INDEXED(cpu,X);mu8=READM(cpu,mu16);M_INS(cpu,mu8);WRITEM(cpu,mu16,mu8); cpu->PC.aword+=3;                /* INS $abcd,X */
  }
