 /*
  * d1551.c - Commodore 1551 floppy drive emulation
  * Copyright (c)2002 Christian Nowak <chnowak@web.de>
  *
  *
  * This file is part of PlusIV.
  *
  * PlusIV is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * PlusIV is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with Foobar; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "d64.h"
#include "d1551.h"
#include "defines.h"

int D1551_Init ( D1551 * drive )
  {
    if (drive==NULL)
      return -1;

    drive->IEC_Data = 0;
    drive->IEC_Status = 0;
    drive->IEC_Handshake = 0;
    drive->IEC_LastCommand = 0;
    drive->IEC_DriveStatus = IEC_IDLE;
    if ( (drive->IEC_InpBuf = malloc(IEC_BUFSIZE))==NULL )
      return -1;
    if ( (drive->IEC_OutBuf = malloc(IEC_BUFSIZE))==NULL )
      {
        free (drive->IEC_InpBuf);
        return -1;
      }
    drive->IEC_OutBufPointer = 0;
    drive->IEC_InpBufPointer = 0;
    drive->IEC_OutBufLength = 0;
    drive->disk.image = NULL;

    return 0;
  }

void D1551_Terminate ( D1551 * drive )
  {
    if (drive==NULL)
      return;

    if (drive->IEC_InpBuf!=NULL)
      free (drive->IEC_InpBuf);
    if (drive->IEC_OutBuf!=NULL)
      free (drive->IEC_OutBuf);

    D64_UnloadImage ( &drive->disk );
  }

void InterpretInputBuffer(D1551 * drive)
  {
    int i;
/*    for (i=0;i<drive->IEC_InpBufPointer;i++)
      fprintf(stderr,"%c",drive->IEC_InpBuf[i]);
    fprintf(stderr,"|\n");*/

    if ( (memcmp(drive->IEC_InpBuf, "$", drive->IEC_InpBufPointer)==0) ||
         (memcmp(drive->IEC_InpBuf, "0:$", drive->IEC_InpBufPointer)==0) )
      { /* Do a directory */
        int s = D64_Directory ( &drive->disk, (char*)drive->IEC_OutBuf );
        if (s==-1)
          {
            drive->IEC_OutBufLength = 0;
            drive->IEC_Status = 3;
          } else
          {
            drive->IEC_OutBufLength = s;
            drive->IEC_Status = 0;
          }
        drive->IEC_OutBufPointer = 0;
      } else
    if ( drive->IEC_InpBufPointer>=3 && memcmp(drive->IEC_InpBuf,"0:",2)==0 )
      { /* Load file */
        char fname[24];

        for (i=0;i<drive->IEC_InpBufPointer;i++)
          fname[i] = drive->IEC_InpBuf[i+2];
        fname[i]='\0';
        drive->IEC_Status = 0;
        if ( (i=D64_LoadFile(&drive->disk,fname,drive->IEC_OutBuf))!=-1 )
          {
            drive->IEC_OutBufLength = i;
          } else
          drive->IEC_Status = 3;
        drive->IEC_OutBufPointer = 0;
      } else
      {
        char fname[24];

        for (i=0;i<drive->IEC_InpBufPointer;i++)
          fname[i] = drive->IEC_InpBuf[i];
        fname[i]='\0';
        drive->IEC_Status = 0;
        if ( (i=D64_LoadFile(&drive->disk,fname,drive->IEC_OutBuf))!=-1 )
          {
            drive->IEC_OutBufLength = i;
          } else
          drive->IEC_Status = 3;
        drive->IEC_OutBufPointer = 0;
      }
    drive->IEC_InpBufPointer = 0;
  }


u8 D1551_IORead(D1551 * drive, u8 a)
  {
    u8 d=0x00;

    switch(a)
      {
        case 0xf0:  /* data */
          d = drive->IEC_Data;
          break;
        case 0xf1:  /* status */
          d = drive->IEC_Status;
          break;
        case 0xf2:  /* handshake */
          d = drive->IEC_Handshake;
          break;
      }
    return d;
  }

void D1551_IOWrite(D1551 * drive, u8 a, u8 d)
  {
    int i;

    switch(a)
      {
        case 0xf0:  /* data */
          switch(drive->IEC_DriveStatus)
            {
              case IEC_IDLE:
                if ( d==0x55 )
                  {
                    drive->IEC_Data = d;
                    drive->IEC_Status = 0;
                  } else
                if ( d==0x81 )  /* Device ID */
                  {
                    drive->IEC_Handshake = 0;
                    drive->IEC_DriveStatus = IEC_DEVICEID;
                  } else
                if ( d==0x82 )  /* Next will be a command (0xe0, 0xf0) */
                  {
                    drive->IEC_Handshake = 0;
                  } else
                if ( d==0x83 )  /* Data to drive */
                  {
                    drive->IEC_DriveStatus = IEC_INPUT;
                  } else
                if ( d==0x84 )  /* Data to Plus/4 */
                  {
                    /* IEC_Data = next byte from floppy */
                    drive->IEC_Handshake = 0;
                    drive->IEC_Data = drive->IEC_OutBuf[drive->IEC_OutBufPointer++];
                    drive->IEC_Status = 0;
                    if ( --drive->IEC_OutBufLength<=0 )
                      {
                        drive->IEC_OutBufPointer = 0;
                        drive->IEC_OutBufLength  = 0;
                        drive->IEC_Status = 3;
                      }
                    /* if (endofbuffer) IEC_Status=3; else IEC_Status=0; */
                  } else
                /* Commands preceding a 0x82 */
                if ( d==0xe0 )
                  { /* Close */
                    drive->IEC_LastCommand = d;
                    drive->IEC_Status = 0;
                    drive->IEC_Handshake = 0xff;
                  }
                break;
              case IEC_DEVICEID:
                if ( d==0x20 )  /* Talk - 1551 gets data */
                  {
                    drive->IEC_InpBufPointer = 0;
                    for (i=0;i<IEC_BUFSIZE;i++)
                      drive->IEC_InpBuf[i] = 0;

                    drive->IEC_Handshake = 0xff;
                    drive->IEC_DriveStatus = IEC_IDLE;
                  } else
                if ( d==0x40 )  /* Plus/4 gets data */
                  {
                    /* Clear output buffer */
                    drive->IEC_Handshake = 0xff;
                    drive->IEC_DriveStatus = IEC_IDLE;
                  } else
                if ( d==0x3f )  /* Untalk */
                  {
                    InterpretInputBuffer(drive);
                    drive->IEC_Handshake = 0xff;
                    drive->IEC_DriveStatus = IEC_IDLE;

                  } else
                if ( d==0x5f )  /* Unlisten */
                  {
                    drive->IEC_Handshake = 0xff;
                    drive->IEC_Status = 0;
                    drive->IEC_DriveStatus = IEC_IDLE;
                  }
                break;
              case IEC_INPUT:
                /* Add byte to input buffer */
                drive->IEC_InpBuf[drive->IEC_InpBufPointer++] = d;
                drive->IEC_DriveStatus = IEC_IDLE;
                break;
            }
        case 0xf1:
          break;
        case 0xf2:
          if ( d==0x00 )
            drive->IEC_Handshake = 0xff;
          else
          if ( d==0x40 )
            drive->IEC_Handshake = 0x00;
          break;
      }
  }
