 /*
  * d64.c - D64 floppy disk image handling
  * Copyright (c)2002 Christian Nowak <chnowak@web.de>
  *
  *
  * This file is part of PlusIV.
  *
  * PlusIV is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * PlusIV is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with Foobar; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "d64.h"

char * d64_ftypes[8] =
  {
    "DEL",
    "SEQ",
    "PRG",
    "USR",
    "REL",
    "???",
    "???",
    "???"
  };

int d64_numsectors[35] =
  {
    /* 1  */  21,
    /* 2  */  21,
    /* 3  */  21,
    /* 4  */  21,
    /* 5  */  21,
    /* 6  */  21,
    /* 7  */  21,
    /* 8  */  21,
    /* 9  */  21,
    /* 10 */  21,
    /* 11 */  21,
    /* 12 */  21,
    /* 13 */  21,
    /* 14 */  21,
    /* 15 */  21,
    /* 16 */  21,
    /* 17 */  21,
    /* 18 */  19,
    /* 19 */  19,
    /* 20 */  19,
    /* 21 */  19,
    /* 22 */  19,
    /* 23 */  19,
    /* 24 */  19,
    /* 25 */  18,
    /* 26 */  18,
    /* 27 */  18,
    /* 28 */  18,
    /* 29 */  18,
    /* 30 */  18,
    /* 31 */  17,
    /* 32 */  17,
    /* 33 */  17,
    /* 34 */  17,
    /* 35 */  17
  };

int d64_trackoffset[35] =
  {
    /* 1  */  0x00000,
    /* 2  */  0x01500,
    /* 3  */  0x02a00,
    /* 4  */  0x03f00,
    /* 5  */  0x05400,
    /* 6  */  0x06900,
    /* 7  */  0x07e00,
    /* 8  */  0x09300,
    /* 9  */  0x0a800,
    /* 10 */  0x0bd00,
    /* 11 */  0x0d200,
    /* 12 */  0x0e700,
    /* 13 */  0x0fc00,
    /* 14 */  0x11100,
    /* 15 */  0x12600,
    /* 16 */  0x13b00,
    /* 17 */  0x15000,
    /* 18 */  0x16500,
    /* 19 */  0x17800,
    /* 20 */  0x18b00,
    /* 21 */  0x19e00,
    /* 22 */  0x1b100,
    /* 23 */  0x1c400,
    /* 24 */  0x1d700,
    /* 25 */  0x1ea00,
    /* 26 */  0x1fc00,
    /* 27 */  0x20e00,
    /* 28 */  0x22000,
    /* 29 */  0x23200,
    /* 30 */  0x24400,
    /* 31 */  0x25600,
    /* 32 */  0x26700,
    /* 33 */  0x27800,
    /* 34 */  0x28900,
    /* 35 */  0x29a00
  };

void stru ( char * s )
  {
    int i,l = strlen(s);

    for (i=0;i<l;i++)
      s[i] = toupper ( s[i] );
  }

D64_BAM * D64_GetBAM ( D64 * disk )
  {
    return ( (D64_BAM*)&disk->image[d64_trackoffset[18-1]] );
  }

int D64_GetFreeBlocks ( D64 * disk )
  {
    int i,free;

    for (i=free=0;i<35;i++)
      free += disk->bam->entries[i].free_sectors;
    free -= disk->bam->entries[18-1].free_sectors;

    return free;
  }

void D64_GetDiskName ( D64 * disk, char * str )
  {
    int i;

    for (i=0;i<16;i++)
      {
        u8 c = disk->bam->disk_name[i];
        if (c==0xa0)
          c=' ';
        str[i] = c;
      }
    str[16] = '\0';
  }

void D64_GetDiskID ( D64 * disk, char * str )
  {
    str[0] = disk->bam->disk_id[0];
    str[1] = disk->bam->disk_id[1];
    str[2] = '\0';
  }

void D64_GetDOSType ( D64 * disk, char * str )
  {
    str[0] = disk->bam->dos_type[0];
    str[1] = disk->bam->dos_type[1];
    str[2] = '\0';
  }

void D64_GetEntryName ( D64 * disk, char * s )
  {
    int i;

    for (i=0;(i<16) && ((u8)disk->direntry->file_name[i]!=0xa0);i++)
      s[i] = disk->direntry->file_name[i];
    s[i] = '\0';
  }

int D64_GetEntrySize ( D64 * disk )
  {
    return disk->direntry->file_size_low | (disk->direntry->file_size_high<<8);
  }

int D64_GotoFirstDirEntry ( D64 * disk )
  {
    disk->curdirtrack = 18;
    disk->curdirsector = 1;
    disk->curdirentry = 0;

    disk->direntry = (D64_direntry*)
      &disk->image[TRACKSECTOR((int)disk->curdirtrack, (int)disk->curdirsector)+
                   ((int)disk->curdirentry<<5)];

    while (disk->direntry->file_type == 0x00)
      {
        if (disk->curdirentry >= 7)
          { /* Go to next track/sector */
            D64_direntry * de = (D64_direntry*) &disk->image[TRACKSECTOR(disk->curdirtrack,disk->curdirsector)];

            disk->curdirentry = 0;
            disk->curdirtrack  = de->next_track;
            disk->curdirsector = de->next_sector;
            if (disk->curdirtrack == 0)
              return -1;
          } else
          disk->curdirentry++;
        disk->direntry = (D64_direntry*)
          &disk->image[TRACKSECTOR((int)disk->curdirtrack, (int)disk->curdirsector)+
                       ((int)disk->curdirentry<<5)];
      }
    return 0;
  }

int D64_GotoNextDirEntry ( D64 * disk )
  {
    do
      {
        if (disk->curdirentry >= 7)
          { /* Go to next track/sector */
            D64_direntry * de = (D64_direntry*)&disk->image[TRACKSECTOR(disk->curdirtrack,disk->curdirsector)];

            disk->curdirentry = 0;
            disk->curdirtrack  = de->next_track;
            disk->curdirsector = de->next_sector;
            if (disk->curdirtrack == 0)
              return -1;
          } else
          disk->curdirentry++;
        disk->direntry = (D64_direntry*)
          &disk->image[TRACKSECTOR((int)disk->curdirtrack, (int)disk->curdirsector)+
                       ((int)disk->curdirentry<<5)];
      } while (disk->direntry->file_type==0x00);
    return 0;
  }

int D64_LoadImage ( D64 * disk, char * fname )
  {
    FILE * fh;
    /* Read d64 image */
    if ( (fh = fopen ( fname, "rb" ))==NULL )
      return 1;

    if ( (disk->image = malloc ( 174848 ))==NULL )
      {
        fclose ( fh );
        return 2;
      }
    if ( fread ( disk->image, 174848, 1, fh )<1 )
      {
        fclose ( fh );
        free ( disk->image );
        disk->image = NULL;
        return 3;
      }
    fclose ( fh );

    disk->bam = D64_GetBAM ( disk );

    return 0;
  }

void D64_UnloadImage ( D64 * disk )
  {
    if (disk==NULL)
      return;
    if (disk->image != NULL)
      {
        free (disk->image);
        disk->image = NULL;
      }
  }

int D64_SeekFile ( D64 * disk, char * fname )
  {
    char s[24],d[24];

    if (disk->image==NULL)
      return -1;

    if (strlen(fname)>16)
      return -1;
    strcpy ( s, fname );
    stru ( s );
    if (D64_GotoFirstDirEntry(disk)!=0)
      return -1;

    D64_GetEntryName ( disk, d );
    stru ( d );
    if (strcmp(s,d)==0)
      return 0;

    while (D64_GotoNextDirEntry(disk)==0)
      {
        D64_GetEntryName ( disk, d );
        stru ( d );
        if (strcmp(s,d)==0)
          return 0;
      }

    return -1;
  }

int D64_SeekFileNum ( D64 * disk, int num )
  {
    int i;

    for (i=0;i!=num;i++)
      {
        int res;
        if (i==0)
          {
            res = D64_GotoFirstDirEntry(disk);
          } else
          {
            res = D64_GotoNextDirEntry(disk);
          }
        if (res!=0)
          return -1;
      }

    return 0;
  }

int D64_LoadFile ( D64 * disk, char * fname, u8 * dest )
  {
    int dofs=0, sofs=0, len=0, size=0;
    int track,sector;

    if (disk->image==NULL)
      return -1;
    if (D64_SeekFile(disk, fname)!=0)
      return -1;

    track  = disk->direntry->first_track_of_file;
    sector = disk->direntry->first_sector_of_file;
    if (track==0)
      return -1;

    while (1)
      {
        sofs   = TRACKSECTOR(track,sector);
        size += len = track==0?sector:254;
        memcpy ( &dest[dofs], &disk->image[sofs+2], len );
        if ( track==0 )
          return size;

        dofs += len;
        track  = disk->image[sofs];
        sector = disk->image[sofs+1];
      }

    return size;
  }

void D64_AppendDirEntry ( D64 * disk, char * out, int * ofs, u16 * addr )
  {
    int blocks,t,i;
    char str[24];

    blocks = D64_GetEntrySize(disk);
    (*addr)+=31;
    out[(*ofs)++] = (*addr)&0xff;
    out[(*ofs)++] = (*addr)>>8;
    out[(*ofs)++] = blocks&0xff;
    out[(*ofs)++] = blocks>>8;
    sprintf(str,"%d",blocks);
    t=strlen(str);
    for (i=0;i<(3-t)+1;i++)
      out[(*ofs)++] = ' ';
    D64_GetEntryName(disk,str);
    out[(*ofs)++] = '\"';
    memcpy ( &out[*ofs], &str, strlen(str) );
    (*ofs)+=strlen(str);
    out[(*ofs)++] = '\"';
    t=strlen(str);
    for (i=0;i<(16-t)+1;i++)
      out[(*ofs)++] = ' ';
    memcpy (&out[*ofs], d64_ftypes[disk->direntry->file_type&7], 3);
    (*ofs)+=3;
    out[(*ofs)++] = 0;
  }

int D64_Directory ( D64 * disk, char * out )
  {
    int ofs = 0;
    char str[24];
    u16 addr = 0x1001;
    int t;

    if (disk->image==NULL)
      return -1;

    out[ofs++] = addr&0xff;
    out[ofs++] = addr>>8;
    addr+=31;
    out[ofs++] = addr&0xff;
    out[ofs++] = addr>>8;
    out[ofs++] = 0;
    out[ofs++] = 0;
    out[ofs++] = 0x12;
    out[ofs++] = '\"';
    D64_GetDiskName(disk, str);
    memcpy ( &out[ofs], &str, 16 );
    ofs+=16;
    out[ofs++] = '\"';
    out[ofs++] = ' ';
    D64_GetDiskID ( disk, str );
    memcpy ( &out[ofs], &str, 2 );
    ofs+=2;
    out[ofs++] = ' ';
    D64_GetDOSType ( disk, str );
    memcpy ( &out[ofs], &str, 2 );
    ofs+=2;
    out[ofs++] = 0;

    if (D64_GotoFirstDirEntry(disk)!=-1)
      {
        D64_AppendDirEntry(disk, out, &ofs, &addr);
        while (D64_GotoNextDirEntry(disk)!=-1)
          {
            D64_AppendDirEntry(disk, out, &ofs, &addr);
          }
      }

    addr+=31;
    out[ofs++] = addr&0xff;
    out[ofs++] = addr>>8;
    t = D64_GetFreeBlocks(disk);
    out[ofs++] = t&0xff;
    out[ofs++] = t>>8;
    out[ofs++] = 'B';
    out[ofs++] = 'L';
    out[ofs++] = 'O';
    out[ofs++] = 'C';
    out[ofs++] = 'K';
    out[ofs++] = 'S';
    out[ofs++] = ' ';
    out[ofs++] = 'F';
    out[ofs++] = 'R';
    out[ofs++] = 'E';
    out[ofs++] = 'E';
    out[ofs++] = 0;
    out[ofs++] = 0;
    out[ofs++] = 0;
    out[ofs++] = 0;
    return ofs;
  }
