 /*
  * d64.h - D64 floppy disk image handling declarations
  * Copyright (c)2002 Christian Nowak <chnowak@web.de>
  *
  *
  * This file is part of PlusIV.
  *
  * PlusIV is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * PlusIV is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with Foobar; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */

#ifndef __D64_H__
#define __D64_H__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "defines.h"

#define TRACKSECTOR(track,sector) (d64_trackoffset[(track)-1]+((sector)<<8))

typedef struct D64_BAM
  {
    u8 dir_track;
    u8 dir_sector;
    u8 version;
    u8 unused1;
    struct entries
      {
        u8 free_sectors;
        u8 free[3];
      } entries[35];
    char disk_name[16];
    u8 a0[2];
    char disk_id[2];
    u8 a02;
    char dos_type[2];
    u8 a03[4];
    u8 unused2[85];
  } D64_BAM;

typedef struct D64_direntry
  {
    u8 next_track;
    u8 next_sector;
    u8 file_type;
    u8 first_track_of_file;
    u8 first_sector_of_file;
    char file_name[16];
    u8 rel_first_track;
    u8 rel_first_sector;
    u8 rel_record_length;
    u8 unused1[6];
    u8 file_size_low;
    u8 file_size_high;
  } D64_direntry;

typedef struct D64
  {
    u8 * image;
    D64_BAM * bam;
    D64_direntry * direntry;
    u16 curdirtrack;
    u16 curdirsector;
    u16 curdirentry;
  } D64;

D64_BAM * D64_GetBAM ( D64 * );
int D64_GetFreeBlocks ( D64 * );
void D64_GetDiskName ( D64 *, char * );
void D64_GetDiskID ( D64 *, char * );
void D64_GetDOSType ( D64 *, char * );
void D64_GetEntryName ( D64 *, char * );
int D64_GetEntrySize ( D64 * );
int D64_GotoFirstDirEntry ( D64 * );
int D64_GotoNextDirEntry ( D64 * );
int D64_LoadImage ( D64 *, char * );
void D64_UnloadImage ( D64 * );
int D64_SeekFile ( D64 *, char * );
int D64_SeekFileNum ( D64 *, int );
int D64_LoadFile ( D64 *, char *, u8 * );
int D64_Directory ( D64 *, char * );

#endif
