 /*
  * main.c - c-style 'main' function for SDL
  * Copyright (c)2002 Christian Nowak <chnowak@web.de>
  *
  *
  * This file is part of PlusIV.
  *
  * PlusIV is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * PlusIV is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with Foobar; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */

#define NDEBUG

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "defines.h"
#include "sdl/sdl.h"
#include "p4/P4.h"
#include "p4/P4Colors.h"

BOOL debug = FALSE;

static void mixaudio ( void * userdata, Uint8 * stream, int len )
  {
    P4 * machine = userdata;

    memcpy ( stream, machine->soundbuf, len );
    P4_RenderSound(machine);
  }

int pollmachine ( P4 * machine )
  {
    SDL_Event event;
    while ( SDL_PollEvent ( &event ) != 0 )
      {
        switch ( event.type )
          {
            case SDL_QUIT:
              return 1;
              break;
            case SDL_KEYDOWN:
            case SDL_KEYUP:
              {
                u8 s = event.type==SDL_KEYDOWN?1:0;
/*                printf("%d\n",event.key.keysym.sym);*/
                switch ( event.key.keysym.sym )
                  {
                    case SDLK_TAB:
/*                      P4_LoadPRG(machine,"_prgs/mine_sweeper.prg");*/
                      P4_LoadPRG(machine,"gp32/gpmm/028/g_man.prg");
                      break;
                    case SDLK_CAPSLOCK:
                      machine->keys[7][7] = s;
                      break;
                    case SDLK_ESCAPE:
                      machine->keys[7][7] = s;
                      if (s==1) debug=!debug;
                      break;
                    case SDLK_0:
                      machine->keys[4][3] = s;
                      break;
                    case SDLK_1:
                      machine->keys[7][0] = s;
                      break;
                    case SDLK_2:
                      machine->keys[7][3] = s;
                      break;
                    case SDLK_3:
                      machine->keys[1][0] = s;
                      break;
                    case SDLK_4:
                      machine->keys[1][3] = s;
                      break;
                    case SDLK_5:
                      machine->keys[2][0] = s;
                      break;
                    case SDLK_6:
                      machine->keys[2][3] = s;
                      break;
                    case SDLK_7:
                      machine->keys[3][0] = s;
                      break;
                    case SDLK_8:
                      machine->keys[3][3] = s;
                      break;
                    case SDLK_9:
                      machine->keys[4][0] = s;
                      break;
                    case SDLK_a:
                      machine->keys[1][2] = s;
                      break;
                    case SDLK_b:
                      machine->keys[3][4] = s;
                      break;
                    case SDLK_c:
                      machine->keys[2][4] = s;
                      break;
                    case SDLK_d:
                      machine->keys[2][2] = s;
                      break;
                    case SDLK_e:
                      machine->keys[1][6] = s;
                      break;
                    case SDLK_f:
                      machine->keys[2][5] = s;
                      break;
                    case SDLK_g:
                      machine->keys[3][2] = s;
                      break;
                    case SDLK_h:
                      machine->keys[3][5] = s;
                      break;
                    case SDLK_i:
                      machine->keys[4][1] = s;
                      break;
                    case SDLK_j:
                      machine->keys[4][2] = s;
                      break;
                    case SDLK_k:
                      machine->keys[4][5] = s;
                      break;
                    case SDLK_l:
                      machine->keys[5][2] = s;
                      break;
                    case SDLK_m:
                      machine->keys[4][4] = s;
                      break;
                    case SDLK_n:
                      machine->keys[4][7] = s;
                      break;
                    case SDLK_o:
                      machine->keys[4][6] = s;
                      break;
                    case SDLK_p:
                      machine->keys[5][1] = s;
                      break;
                    case SDLK_q:
                      machine->keys[7][6] = s;
                      break;
                    case SDLK_r:
                      machine->keys[2][1] = s;
                      break;
                    case SDLK_s:
                      machine->keys[1][5] = s;
                      break;
                    case SDLK_t:
                      machine->keys[2][6] = s;
                      break;
                    case SDLK_u:
                      machine->keys[3][6] = s;
                      break;
                    case SDLK_v:
                      machine->keys[3][7] = s;
                      break;
                    case SDLK_w:
                      machine->keys[1][1] = s;
                      break;
                    case SDLK_x:
                      machine->keys[2][7] = s;
                      break;
                    case SDLK_y:
                      machine->keys[3][1] = s;
                      break;
                    case SDLK_z:
                      machine->keys[1][4] = s;
                      break;
                    case SDLK_UP:
                      machine->keys[5][3] = s;
                      break;
                    case SDLK_DOWN:
                      machine->keys[5][0] = s;
                      break;
                    case SDLK_LEFT:
                      machine->keys[6][0] = s;
                      break;
                    case SDLK_RIGHT:
                      machine->keys[6][3] = s;
                      break;
                    case SDLK_RETURN:
                      machine->keys[0][1] = s;
                      break;
                    case SDLK_SPACE:
                      machine->keys[7][4] = s;
                      break;
                    case SDLK_BACKSPACE:
                      machine->keys[0][0] = s;
                      break;
                    case SDLK_EQUALS:
                      machine->keys[6][5] = s;
                      break;
                    case SDLK_COMMA:
                      machine->keys[5][7] = s;
                      break;
                    case SDLK_PERIOD:
                      machine->keys[5][4] = s;
                      break;
                    case SDLK_RIGHTBRACKET: /* Plus */
                      machine->keys[6][6] = s;
                      break;
                    case 59:  /* Minus */
                      machine->keys[5][6] = s;
                      break;
                    case 39: /* Asterisk */
                      machine->keys[6][1] = s;
                      break;
                    case SDLK_F1:
                      machine->keys[0][4] = s;
                      break;
                    case SDLK_F2:
                      machine->keys[0][5] = s;
                      break;
                    case SDLK_F3:
                      machine->keys[0][6] = s;
                      break;
                    case SDLK_RSHIFT:
                    case SDLK_LSHIFT:
                      machine->keys[1][7] = s;
                      break;
                    case SDLK_BACKQUOTE:  /* ESC */
                      machine->keys[6][4] = s;
                      break;
                    case SDLK_KP8:  /* Joy Up */
                      machine->joy1[0] = s;
                      break;
                    case SDLK_KP2:  /* Joy Down */
                      machine->joy1[1] = s;
                      break;
                    case SDLK_KP4:  /* Joy Left */
                      machine->joy1[2] = s;
                      break;
                    case SDLK_KP6:  /* Joy Right */
                      machine->joy1[3] = s;
                      break;
                    case SDLK_KP0:  /* Joy Fire */
                      machine->joy1[4] = s;
                      break;
                    case SDLK_KP7:
                      if (s==1)
                        P4_Reset ( machine );
                      break;
                    default:
                      break;
                  }
                break;
              }
          }
      }
    return 0;
  }

int main ( void )
  {
    P4 machine;
/*    u8 * m;
    FILE * fh;*/
    int i, frm=0, tick,tick2;
    SDL_Surface *sbuf1,*sbuf2;
    SDL_Color colors[256];
    SDL_AudioSpec audioSpec;
    BOOL audioPlaying = FALSE;
    FILE * f;

    assert (sizeof(u8)==1);
    assert (sizeof(s8)==1);
    assert (sizeof(u16)==2);
    assert (sizeof(s16)==2);
    assert (sizeof(u32)==4);
    assert (sizeof(s32)==4);

    if ( SDL_Init ( SDL_INIT_VIDEO|SDL_INIT_AUDIO ) < 0 )
      {
        fprintf(stderr,"Couldn't initialize SDL\n");
        return 1;
      }
    atexit ( SDL_Quit );

    /* Init SDL video */
    sbuf1 = SDL_SetVideoMode ( 320, 200, 8, (SDL_GetVideoInfo()->hw_available?SDL_HWSURFACE:SDL_SWSURFACE)|SDL_HWPALETTE );
    if ( sbuf1 == NULL )
      {
        fprintf ( stderr, "Couldn't set video mode: %s\n", SDL_GetError() );
        return 1;
      }

    sbuf2 = SDL_DisplayFormat(sbuf1);
    SDL_WM_SetCaption ( "P4emu by CHN", NULL );
    for (i=0;i<=128;i++)
      {
        u8 c = ((i&0x0f)<<3)|((i>>4)&0x07);
        colors[i].r = P4Colors[c][0];
        colors[i].g = P4Colors[c][1];
        colors[i].b = P4Colors[c][2];
      }
    SDL_SetPalette ( sbuf1, SDL_PHYSPAL, colors, 0, 256 );

    /* Init emulation */
    if (!P4_Init ( &machine ))
      {
        printf("Couldn't init machine\n");
        return 1;
      }

    /* Init SDL audio */
    audioSpec.freq = 22050;
    audioSpec.format = AUDIO_U16;
    audioSpec.channels = 1;
    audioSpec.samples = machine.soundbufsize;
    audioSpec.callback = mixaudio;
    audioSpec.userdata = (void*)&machine;
    if (SDL_OpenAudio(&audioSpec,NULL)<0)
      {
        fprintf(stderr,"Error while opening audio device\n");
      }
    else
      {
        audioPlaying = TRUE;
        SDL_PauseAudio(0);
      }

    D64_LoadImage ( &machine.drive->disk, "_prgs\\terranov.d64" );
/*    m = calloc ( 1, 65536 );
    if ( (i=D64_LoadFile ( &machine.drive->disk, "TERRA NOVA +4", m ))>0 )
      {
        fh = fopen ( "terranova.prg", "wb" );
        if (fh!=NULL)
          {
            fwrite ( m, 1, i, fh );
            fclose ( fh );
          }
      }
    free ( m );*/
    
    tick = SDL_GetTicks();
    while(1)
      {
        u16 curline = machine.rasterCounter;
        /* Render line here */
        if ( SDL_MUSTLOCK(sbuf2) )
          SDL_LockSurface ( sbuf2 );
        machine.videoBuf = sbuf2->pixels;
        P4_Run (&machine);
        if ( SDL_MUSTLOCK(sbuf2) )
          SDL_UnlockSurface ( sbuf2 );

        if (machine.rasterCounter==0 && curline>200)
          {
            SDL_BlitSurface( sbuf2, NULL , sbuf1, NULL);
            SDL_Flip(sbuf1);
            SDL_FillRect ( sbuf2, NULL, 0 );
            if(pollmachine(&machine)) break;
            frm++;
            tick2 = SDL_GetTicks();
            while (((tick2=SDL_GetTicks())-tick)<20);
            tick=tick2;
          }
      }
/*
    f = fopen("ram.bin","wb");
    if (f!=NULL)
      {
        fwrite(machine.RAM, 64*1024, 1, f);
        fclose(f);
      }*/

    if (audioPlaying)
      SDL_CloseAudio();

    P4_Terminate ( &machine );

    return 0;
  }
