 /*
  * P4.c - Plus/4 TED emulation (graphics, sound, memory mapping, etc.)
  * Copyright (c)2002 Christian Nowak <chnowak@web.de>
  *
  *
  * This file is part of PlusIV.
  *
  * PlusIV is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * PlusIV is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with Foobar; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "P4.h"
#include "../cpu/M6502.h"
#include "defines.h"
#include "freqtab.h"
#include "noise.h"
#include "../roms/kernal.h"
#include "../roms/basic.h"
#include "../roms/3plus1lo.h"
#include "../roms/3plus1hi.h"
#include "../disk/d1551.h"

u16 volumetab[16] =
  {
    0,
    2047>>1,
    4095>>1,
    6143>>1,
    8191>>1,
    10239>>1,
    12287>>1,
    14335>>1,
    16383>>1,
    16383>>1,
    16383>>1,
    16383>>1,
    16383>>1,
    16383>>1,
    16383>>1,
    16383>>1
  };

extern BOOL debug;

#define XCHG8(a,b)  {u8 tmp = b;b=a;a=tmp;}
#define readLow(m,a)  (m->low!=NULL?m->low[a&16383]:0x00)
#define readHigh(m,a) (m->high!=NULL?m->high[a&16383]:0x00)
#define TEDREAD(m,a)  ((m->regs[0x12]&0x04)?(((a)&0x4000)?readHigh(m,(a)):readLow(m,(a))):m->RAM[a])
#define	CLOCKSPERLINE	(((machine->regs[0x13]&2)==0)?(machine->rasterCounter<204?67:113):57)
static int fsize ( char * fname )
  {
    FILE * f;
    int s;

    f = fopen (fname,"rb");
    if (f==NULL)
      return -1;

    fseek(f, 0, SEEK_END);
    s=ftell(f);
    fclose(f);

    return s;
  }

void TED_RegWrite ( P4 * machine, u8 r, u8 d )
  {
    switch (r)
      {
        case 0x00:  /* Timer 1 count low */
          machine->timer1inhibit=TRUE;
          machine->timer1reload = (machine->timer1reload&0xff00) | d;
          break;
        case 0x01:  /* Timer 1 count high */
          machine->timer1inhibit=FALSE;
          machine->timer1reload = (machine->timer1reload&0x00ff) | (d<<8);
          break;
        case 0x02:  /* Timer 2 count low */
          machine->timer2inhibit=TRUE;
          machine->timer2count  = (machine->timer2count&0xff00) | d;
          break;
        case 0x03:  /* Timer 2 count high */
          machine->timer2inhibit=FALSE;
          machine->timer2count  = (machine->timer2count&0x00ff) | (d<<8);
          break;
        case 0x04:  /* Timer 3 count low */
          machine->timer3inhibit=TRUE;
          machine->timer3count  = (machine->timer3count&0xff00) | d;
          break;
        case 0x05:  /* Timer 3 count high */
          machine->timer3inhibit=FALSE;
          machine->timer3count  = (machine->timer3count&0x00ff) | (d<<8);
          break;
        case 0x0a:
          machine->rasterCounterCompare =
            (machine->rasterCounterCompare & 0x00ff) | (((u16)d<<8)&0x0100);
          break;
        case 0x0b:
          machine->rasterCounterCompare =
            (machine->rasterCounterCompare & 0x0100) | (((u16)d)&0x00ff);
          break;
        case 0x13:
          machine->clock = (d&0x02)==0;
/*          printf("Write character base %02x %d %04x\n",d,machine->rasterCounter,((M6502*)machine->cpu)->PC.word);*/
          break;
        case 0x1d:
          machine->rasteryofs=0;
/*          machine->rasterCounter=(machine->rasterCounter&0x100)|d;*/
/*         	printf("%d -> %d\n",machine->rasterCounter, d);*/
/*          machine->rasterCounter = d;*/
          break;
        case 0x3e:
          machine->ramBankedIn=FALSE;
          break;
        case 0x3f:
          machine->ramBankedIn=TRUE;
          break;
        case 0x09:
        	if (d&0x80)
          	machine->IRQ = FALSE;
          if (d&0x08)
            machine->timer1IRQ = FALSE;
          if (d&0x10)
            machine->timer2IRQ = FALSE;
          if (d&0x40)
            machine->timer3IRQ = FALSE;
          if (d&0x02)
            machine->rasterIRQ = FALSE;
          break;
        default:
          break;
      }
    machine->regs[r] = d;
    machine->RAM[0xff00|r] = d;
  }

u8 TED_RegRead ( P4 * machine, u8 r )
  {
    switch (r)
      {
        case 0x00:  /* Timer 1 count value low */
          return (machine->timer1count&0xff);
          break;
        case 0x01:  /* Timer 1 count value high */
          return ((machine->timer1count>>8)&0xff);
          break;
        case 0x02:  /* Timer 2 count value low */
          return (machine->timer2count&0xff);
          break;
        case 0x03:  /* Timer 2 count value high */
          return ((machine->timer2count>>8)&0xff);
          break;
        case 0x04:  /* Timer 3 count value low */
          return (machine->timer3count&0xff);
          break;
        case 0x05:  /* Timer 3 count value high */
          return ((machine->timer3count>>8)&0xff);
          break;
        case 0x13:
          return (machine->ramBankedIn?0:1) | (machine->regs[0x13]&0xfc);
        case 0x08:
            {
              u8 retval = 0xff;
              int i,j;

              for (i=0;i<8;i++)
                if (~machine->keylatch & (1<<i))
                  {
                    for (j=0;j<8;j++)
                      {
                        if (machine->keys[i][j]==1)
                          retval &= ~(1<<j);
                      }
                  }

              /* Joystick 1 */
              if (~machine->regs[8]&(1<<2))
                {
                  if (machine->joy1[0]==1)
                    retval&=0xfe;
                  if (machine->joy1[1]==1)
                    retval&=0xfd;
                  if (machine->joy1[2]==1)
                    retval&=0xfb;
                  if (machine->joy1[3]==1)
                    retval&=0xf7;
                  if (machine->joy1[4]==1)
                    retval&=0xbf;
                }
              /* Joystick 2 */
              if (~machine->regs[8]&(1<<1))
                {
                  if (machine->joy2[0]==1)
                    retval&=0xfe;
                  if (machine->joy2[1]==1)
                    retval&=0xfd;
                  if (machine->joy2[2]==1)
                    retval&=0xfb;
                  if (machine->joy2[3]==1)
                    retval&=0xf7;
                  if (machine->joy2[4]==1)
                    retval&=0x7f;
                }
              return retval;
            }
          return 0xff;
        case 0x09:
        	return (machine->IRQ?0x80:0x00) |
                         (machine->timer1IRQ?0x08:0x00) |
                         (machine->timer2IRQ?0x10:0x00) |
                         (machine->timer3IRQ?0x40:0x00) |
                         (machine->rasterIRQ?0x02:0x00) | 0x21;
        case 0x1c:
          return (machine->rasterCounter>>8)&1;
        case 0x1d:
          return machine->rasterCounter&0xff;
        case 0x1e: /* Horiz Scan */
          return (((M6502*)machine->cpu)->cycles*228)/CLOCKSPERLINE;
        case 0x1f:
          return ((machine->flash&0x0f)<<3);
        default:
          return machine->regs[r];
      }
  }

void TED_IOWrite ( P4 * machine, u8 a, u8 d )
  {
    if ((a&0xf0) == 0xd0)
      { /* Banking */
        machine->loBank = a&0x03;
        machine->hiBank = (a>>2)&0x03;
        switch (machine->loBank)
          {
            case 0:
              machine->low = machine->basic;
              break;
            case 1:
              machine->low = machine->functionLo;
              break;
            default:
              machine->low = NULL;
              break;
          }
        switch (machine->hiBank)
          {
            case 0:
              machine->high = machine->kernal;
              break;
            case 1:
              machine->high = machine->functionHi;
              break;
            default:
              machine->high = NULL;
              break;
          }
      } else
    if (a==0x30)
      {
        machine->keylatch = d;
      }
  }


u8 readMem ( M6502 * cpu, u16 a )
  {
    P4 * machine = cpu->val;
    u8 d = 0;

    if ( a<0x8000 )
      {
        d = machine->RAM[a];
      } else
    if ( a<0xc000 )
      {
        /* Low Bank */
        d = (machine->ramBankedIn ? machine->RAM[a] : readLow(machine,a));
      } else
    if ( a<0xfc00 ) /* c000..fbff */
      {
        d = (machine->ramBankedIn ? machine->RAM[a] : readHigh(machine,a));
      } else
    if ( a<0xfd00 ) /* fc00..fcff */
      {
        d = (machine->ramBankedIn ? machine->RAM[a] : machine->kernal[a&16383]);
      } else
    if ( a<0xfe00 ) /* fd00..fdff */
      {
        d = (machine->ramBankedIn ? machine->RAM[a] : readHigh(machine,a));
      } else
    if ( a<0xff00 ) /* fe00..feff */
      {
        d = D1551_IORead ( machine->drive, a&0xff );
      } else
    if ( a<0xff40 ) /* ff00..ff3f */
      {
        /* TED Registers */
        d = TED_RegRead ( machine, a&0x3f );
      } else
      {
        /* a>=0xff40 */
        d = (machine->ramBankedIn ? machine->RAM[a] : readHigh(machine,a));
      }

    return d;
  }

void writeMem ( M6502 * cpu, u16 a, u8 d )
  {
    P4 * machine = cpu->val;

    if ( a<0xfd00 )
      {
          machine->RAM[a]=d;
      } else
    if ( a<0xfe00 )
      {
        TED_IOWrite ( machine, a&0xff, d );
      } else
    if ( a<0xff00 )
      {
        D1551_IOWrite(machine->drive, a&0xff,d);
      } else
    if ( a<0xff40 )
      {
        TED_RegWrite (machine,a&0x3f,d);
      } else
      {
/*        if (machine->ramBankedIn)*/
          machine->RAM[a] = d;
      }
  }

void P4_Terminate ( P4 * machine )
  {
    if (machine==NULL)
      return;
    if ( machine->cpu!=NULL )
      free (machine->cpu);
    if ( machine->RAM!=NULL )
      free (machine->RAM);
    if ( machine->soundbuf != NULL )
      free (machine->soundbuf);
    if ( machine->drive!=NULL )
      {
        D1551_Terminate ( machine->drive );
        free ( machine->drive );
      }
  }

BOOL P4_Init ( P4 * machine )
  {
    D1551 * drive;
    M6502* cpu = calloc ( 1,sizeof(M6502) );
    if (cpu==NULL)
      return FALSE;

    drive = calloc ( 1, sizeof(D1551) );
    if (drive==NULL)
      {
        free(cpu);
        return FALSE;
      }
    memset ( machine, 0, sizeof(P4) );

    if ( (machine->RAM = malloc(64*1024))==NULL )
      {
        P4_Terminate ( machine );
        return FALSE;
      }

    machine->cpu = cpu;
    machine->drive = drive;
    machine->hiBank = 0;
    machine->loBank = 0;
    machine->ramBankedIn = FALSE;
    machine->kernal = kernal;
    machine->basic = basic;
    machine->functionLo = functionlo;
    machine->functionHi = functionhi;
    machine->cartridge = NULL;
    machine->clock = TRUE;
    machine->cycles = 0;
    machine->cyclesPerLine = 113;
    machine->rasterCounterCompare = 0;
    machine->regs[10] = 0x5a;
    machine->low = basic;
    machine->high = kernal;
    machine->cpu->val = machine;

    machine->soundbufsize = 441;
    machine->soundbuf = calloc ( sizeof(u16),machine->soundbufsize );
    if (machine->soundbuf==NULL)
      machine->soundbufsize = 0;

    machine->timer1reload = 0xffff;

    M6502_Init ( cpu, readMem, writeMem );
    D1551_Init ( drive );

    return TRUE;
  }

void P4_Run ( P4 * machine )
  {
    u16 cpusp = CLOCKSPERLINE;
                                      /* Twice the amount of clock cycles per
                                         rasterline on the higher CPU speed */
    u16 c;

    P4_RenderLine ( machine );
    c = M6502_Run ( machine->cpu, cpusp );
/*    ((M6502*)machine->cpu)->cycles -= cpusp;*/
    machine->cycles += c;

    /* Timers */
    if ( machine->clock ) c>>=1;   /* On the faster CPU clock, the timer
                                         IRQs occur in longer intervals (from
                                         the CPU's point of view) */

    if (!machine->timer1inhibit)
      {
        if ( c>=machine->timer1count )
          {
            machine->timer1count -= machine->timer1reload;
            machine->timer1IRQ = TRUE;
            if (machine->regs[10]&0x08)
              {
                machine->IRQ=TRUE;
                ((M6502*)machine->cpu)->IRQB = TRUE;
              }
          }
        machine->timer1count -= c;
      }

    if (!machine->timer2inhibit)
      {
        if ( c>=machine->timer2count )
          {
            machine->timer2IRQ = TRUE;
            if (machine->regs[10]&0x10)
              {
                machine->IRQ=TRUE;
                ((M6502*)machine->cpu)->IRQB = TRUE;
              }
          }
        machine->timer2count -= c;
      }

    if (!machine->timer3inhibit)
      {
        if ( c>=machine->timer3count )
          {
            machine->timer3IRQ = TRUE;
            if (machine->regs[10]&0x40)
              {
                machine->IRQ=TRUE;
                ((M6502*)machine->cpu)->IRQB = TRUE;
              }
          }
        machine->timer3count -= c;
      }

    machine->cycles -= cpusp;

    /* Raster counter & IRQ */
    machine->rasterCounter++;
    if (machine->rasterCounter>311)
      {
        machine->rasterCounter=0;
        machine->flash++;
        machine->flash &= 0x1f;
      }
    if ( machine->rasterCounter == machine->rasterCounterCompare )
      {
        machine->rasterIRQ = TRUE;
        if (machine->regs[10]&0x02)
          {
            machine->IRQ=TRUE;
            ((M6502*)machine->cpu)->IRQB = TRUE;
          }
      }
  }


void P4_RenderSound ( P4 * machine )
  {
    int i;
    u16 snd1freq = machine->regs[0x0e] | ((machine->regs[0x12]<<8)&0x0300);
    u16 snd2freq = machine->regs[0x0f] | ((machine->regs[0x10]<<8)&0x0300);
    s16 v = machine->regs[0x11]&0x0f;
    u16 vol = volumetab[v];
    BOOL s1square = (machine->regs[0x11]&0x10)==0x10;
    BOOL s2square = (machine->regs[0x11]&0x20)==0x20;
    BOOL s2noise = (machine->regs[0x11]&0x40)==0x40;

#define DOSND1\
  machine->snd1count += freqtab[snd1freq];\
  if (machine->snd1count&0xffff0000)\
    {\
      machine->snd1count &= 0x0000ffff;\
      machine->snd1sq = !machine->snd1sq;\
    }
#define DOSND2\
  machine->snd2count += freqtab[snd2freq];\
  if (machine->snd2count&0xffff0000)\
    {\
      machine->snd2count &= 0x0000ffff;\
      machine->snd2sq = !machine->snd2sq;\
    }

    if (v>8) v=8;

    if ( (machine->soundbuf==NULL) || (machine->soundbufsize==0) )
      return;

    if ((!s1square && !s2square && !s2noise) || v==0)
      for (i=0;i<machine->soundbufsize;i++)
        machine->soundbuf[i] =  32768;
    else
    if (s1square && !s2square && !s2noise)
      for (i=0;i<machine->soundbufsize;i++)
        {
          machine->soundbuf[i] = (machine->snd1sq ? vol : -vol) + 32768;
          DOSND1;
        }
    else
    if (s1square && s2square)
      for (i=0;i<machine->soundbufsize;i++)
        {
          machine->soundbuf[i] = (machine->snd1sq ? vol : -vol) +
                                 (machine->snd2sq ? vol : -vol) + 32768;
          DOSND1;
          DOSND2;
        }
    else
    if (s1square && !s2square && s2noise)
      for (i=0;i<machine->soundbufsize;i++)
        {
          machine->soundbuf[i] = (machine->snd1sq ? vol : -vol) +
                                 (((s16)noise[(machine->snd2count>>16)&0x7fff]*(v))>>3) + 32768;
          machine->snd2count += freqtab[snd2freq];
          DOSND1;
        }
    else
    if (!s1square && !s2square && s2noise)
      for (i=0;i<machine->soundbufsize;i++)
        {
          machine->soundbuf[i] = (((s16)noise[(machine->snd2count>>16)&0x7fff]*(v))>>3) + 32768;
          machine->snd2count += freqtab[snd2freq];
        }
    else
    if (!s1square && s2square)
      for (i=0;i<machine->soundbufsize;i++)
        {
          machine->soundbuf[i] = (machine->snd2sq ? vol : -vol) + 32768;
          DOSND2;
        }
  }

void P4_RenderLine ( P4 * m )
  {
	  P4 * machine = m;
    u16 curline = machine->rasterCounter;
    u16 currow;
    u16 curcol;
    u16 curpos;
    u16 cursorpos = (u16)machine->regs[0x0d] | (((u16)machine->regs[0x0c]<<8)&0x0300);
    int xofs = machine->regs[7]&7;
    int yofs = (3-(int)(machine->regs[6]&7))+machine->rasteryofs;
    int temp = curline-4-yofs;
    u8 * videoBuf = &machine->videoBuf[temp*320];
    u16 textstart = (machine->regs[0x14]<<8)&0xf800;
    u16 charbase = ((u16)machine->regs[0x13]<<8) & 0xfc00;
    int i;

    if ( !(temp>=0 && temp<200) )
      return;
    curline -= 4;
    curpos = (curline>>3)*40;

    currow = ((curline>>3)*40);

    /* MCM and ECM at once ? */
    if ( ((machine->regs[0x07]&0x10) && (machine->regs[0x06]&0x40)) || (!(machine->regs[0x06]&0x10)) )
      {
        u16 bgcolor=machine->regs[0x15]&0x7f;
        for (i=0;i<320;i++)
          machine->videoBuf[i] = bgcolor;
      }
    else
    if ( machine->regs[0x06]&0x20 )
      { /* Bitplane Mode */
        u16 bpbase = (machine->regs[0x12]<<10)&0xe000;

        for (curcol=0;curcol<40;curcol++,curpos++)
          {
            u16 curx = xofs+(curcol<<3);
            u8 fgcolor,bgcolor;
            u16 cofs = curcol + currow;
            u8 c = machine->RAM[(textstart+cofs)|0x400];  /* Character # */
            u8 ca= machine->RAM[textstart+cofs];  /* Attribute */
            u8 d = machine->RAM[bpbase+(curcol<<3)+(curline&7)+((curline&~7)*40)];

            fgcolor = ((c>>4)&0x0f)|((ca<<4)&0x70);
            bgcolor = (c&0x0f)|(ca&0x70);

            if (machine->regs[0x07]&0x10)
              { /* Bitplane Multi Color Mode */
                u8 pal[4];

                pal[0] = machine->regs[0x15];
                pal[1] = fgcolor;
                pal[2] = bgcolor;
                pal[3] = machine->regs[0x16];

                videoBuf[0+curx] = videoBuf[1+curx] = pal[(d>>6)&3]&0x7f;
                videoBuf[2+curx] = videoBuf[3+curx] = pal[(d>>4)&3]&0x7f;
                videoBuf[4+curx] = videoBuf[5+curx] = pal[(d>>2)&3]&0x7f;
                videoBuf[6+curx] = videoBuf[7+curx] = pal[(d>>0)&3]&0x7f;
              } else
              {
                videoBuf[0+curx] = d&0x80?fgcolor:bgcolor;
                videoBuf[1+curx] = d&0x40?fgcolor:bgcolor;
                videoBuf[2+curx] = d&0x20?fgcolor:bgcolor;
                videoBuf[3+curx] = d&0x10?fgcolor:bgcolor;
                videoBuf[4+curx] = d&0x08?fgcolor:bgcolor;
                videoBuf[5+curx] = d&0x04?fgcolor:bgcolor;
                videoBuf[6+curx] = d&0x02?fgcolor:bgcolor;
                videoBuf[7+curx] = d&0x01?fgcolor:bgcolor;
              }
          }
      } else
    for (curcol=0;curcol<40;curcol++,curpos++)
      { /* Normal Text Mode */
        u16 cofs = curcol + currow;
        u8 c =machine->RAM[(textstart+cofs)|0x400]; /* Character # */
        u8 ca=machine->RAM[textstart+cofs]; /* Attribute */
        u8 fgcolor=ca&0x7f;
        u8 cl;
        u16 bgcolor=machine->regs[0x15]&0x7f;
        u16 curx = xofs+(curcol<<3);
        u16 tempa;

    		if (machine->regs[0x07]&0x80)
       		charbase&=0xf800;

        if ( (machine->regs[0x07]&0x10) && (ca&0x08) )
          { /* Text Multi Color Mode */
            tempa = charbase+((u16)c<<3)+(curline&7);
            cl=TEDREAD(machine,tempa);

            videoBuf[0+curx] = videoBuf[1+curx] = (cl&0xc0)==0xc0 ? fgcolor&~0x08 : machine->regs[0x15+((cl>>6)&3)]&0x7f;
            videoBuf[2+curx] = videoBuf[3+curx] = (cl&0x30)==0x30 ? fgcolor&~0x08 : machine->regs[0x15+((cl>>4)&3)]&0x7f;
            videoBuf[4+curx] = videoBuf[5+curx] = (cl&0x0c)==0x0c ? fgcolor&~0x08 : machine->regs[0x15+((cl>>2)&3)]&0x7f;
            videoBuf[6+curx] = videoBuf[7+curx] = (cl&0x03)==0x03 ? fgcolor&~0x08 : machine->regs[0x15+((cl>>0)&3)]&0x7f;
          } else
          {
            if ( curpos==cursorpos )
              if ( !(machine->flash&0x10) && !(machine->regs[0x07]&0x10) )
                XCHG8(fgcolor,bgcolor);

            /* ECM */
            if (machine->regs[0x06]&0x40)
              {
                bgcolor = machine->regs[0x15+((c>>6)&3)] & 0x7f;
                c &= 0x3f;
              } else
            /* Reverse */
            if (!(machine->regs[0x07]&0x80))
              {
	              if (c&0x80)
                	XCHG8(bgcolor,fgcolor);
                c&=0x7f;
              }

            tempa = charbase|((u16)c<<3)|(curline&7);
            cl=TEDREAD(machine,tempa);

            videoBuf[0+curx] = cl&0x80?fgcolor:bgcolor;
            videoBuf[1+curx] = cl&0x40?fgcolor:bgcolor;
            videoBuf[2+curx] = cl&0x20?fgcolor:bgcolor;
            videoBuf[3+curx] = cl&0x10?fgcolor:bgcolor;
            videoBuf[4+curx] = cl&0x08?fgcolor:bgcolor;
            videoBuf[5+curx] = cl&0x04?fgcolor:bgcolor;
            videoBuf[6+curx] = cl&0x02?fgcolor:bgcolor;
            videoBuf[7+curx] = cl&0x01?fgcolor:bgcolor;
          }
      }
  }

int P4_LoadPRG ( P4 * machine, char * fname )
  {
    FILE * f;
    int s = -1;

    u16 loc = 0;
    u8 t = 0;

    if ((s = fsize(fname))<0)
      return 1;
    if (s<3)
      return 2;
    if ((f = fopen(fname,"rb"))==NULL)
      return 1;

    s -= 2;
    fread ( &t, 1, 1, f );
    loc = (u16)t;
    fread ( &t, 1, 1, f );
    loc |= (u16)t << 8;

    if ( ((u32)loc+(u32)s)>=65536 )
      {
        fclose (f);
        return 3;
      }

    fread ( &machine->RAM[loc], s, 1, f );
    fclose (f);

    loc = machine->RAM[0x2b] | (machine->RAM[0x2c]<<8);
    loc += s;
    machine->RAM[0x2d] = loc&0xff;
    machine->RAM[0x2e] = (loc>>8)&0xff;
    machine->RAM[0x2f] = machine->RAM[0x2d];
    machine->RAM[0x30] = machine->RAM[0x2e];
    machine->RAM[0x31] = machine->RAM[0x2d];
    machine->RAM[0x32] = machine->RAM[0x2e];
    machine->RAM[0x9d] = machine->RAM[0x2d];
    machine->RAM[0x9e] = machine->RAM[0x2e];

    return 0;
  }

void P4_StartPRG ( P4 * machine )
  {
    if (machine==NULL)
      return;

    machine->RAM[0x00ef] = 0x04;
    machine->RAM[0x0527] = 82;
    machine->RAM[0x0528] = 85;
    machine->RAM[0x0529] = 78;
    machine->RAM[0x052a] = 13;
  }

void P4_Reset ( P4 * machine )
  {
    if (machine==NULL)
      return;

    machine->hiBank = 0;
    machine->loBank = 0;
    machine->ramBankedIn = FALSE;
    machine->clock = TRUE;
    machine->cycles = 0;
    machine->cyclesPerLine = 113;
    machine->rasterCounter = 0;
    machine->rasterCounterCompare = 0;
    machine->regs[10] = 0x5a;
    machine->low = basic;
    machine->high = kernal;
    if (machine->RAM!=NULL)
      memset ( machine->RAM, 0, 65536 );
    memset ( machine->regs, 0, 64 );
    machine->timer1IRQ = FALSE;
    machine->timer2IRQ = FALSE;
    machine->timer3IRQ = FALSE;
    machine->rasterIRQ = FALSE;
    M6502_Reset ( machine->cpu );
  }
