 /*
  * P4.h - Plus/4 TED emulation declarations
  * Copyright (c)2002 Christian Nowak <chnowak@web.de>
  *
  *
  * This file is part of PlusIV.
  *
  * PlusIV is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * PlusIV is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with Foobar; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */

#ifndef __P4_H__
#define __P4_H__

#include "defines.h"
#include "../disk/D1551.h"
#include "../cpu/M6502.h"

typedef struct P4
  {
    M6502 * cpu;
    D1551 * drive;
    u8 hiBank;
    u8 loBank;
    BOOL ramBankedIn;
    u8 * RAM;
    u8 * kernal;    /* Kernal ROM */
    u8 * basic;     /* Basic ROM */
    u8 * cartridge; /* Cartridge ROM */
    u8 * functionLo;  /* Function ROM */
    u8 * functionHi;  /* Function ROM */
    BOOL IRQ;       /* reg 9 bit 7 */
    BOOL timer1IRQ;
    BOOL timer2IRQ;
    BOOL timer3IRQ;
    BOOL rasterIRQ;
    u16 rasterCounter;
    u16 rasterCounterCompare;
    u16 cyclesPerLine;
    u8 regs[64];
    BOOL clock;     /* TRUE=1.768MHz FALSE=884KHz */
    u16 cycles;
    u8 * low;
    u8 * high;
    u8 * videoBuf;  /* 320x200x256 Video Buffer */
    u8 keys[8][8];
    u8 joy1[5];
    u8 joy2[5];
    u8 keylatch;
    u16 flash;
    u16 timer1count;
    u16 timer2count;
    u16 timer3count;
    u16 timer1reload;
    BOOL timer1inhibit;
    BOOL timer2inhibit;
    BOOL timer3inhibit;
    int rasteryofs;
    u16 * soundbuf;   /* Sound buffer - size is 22050Hz/50Hz=441 words */
    int soundbufsize; /* In words */
    u32 snd1count;
    u32 snd2count;
    BOOL snd1sq;
    BOOL snd2sq;
  } P4;

void P4_Reset ( P4 * );
void P4_Run ( P4 * );
BOOL P4_Init ( P4 * );
void P4_Terminate ( P4 * );
void P4_RenderLine ( P4 * );
int P4_LoadPRG ( P4 *, char * );
void P4_StartPRG ( P4 * );
void P4_RenderSound ( P4 * );
#endif
